/*-----------------------------------------------------------------------------
 * geofeather.c
 *
 * demonstrates how to build a feather object in a geometry shader using hair
 * primitives, with a callback.
 *
 * (C) mental images 2003
 *---------------------------------------------------------------------------*/

#include "shader.h"
#include "geoshader.h"


typedef struct
{
	int     hair_number;
	int		shadow_mode;
}geofeather_t;

extern "C"
{

DLLEXPORT miBoolean geofeather_cb(miTag tag, void *args);

DLLEXPORT miInteger geofeather_version(miState *state, geofeather_t *paras, miBoolean *inst_init_req)
{
    return 1;
}


DLLEXPORT miBoolean geofeather(miTag *result, miState *state, geofeather_t *paras)
{
    miTag     tag;
    miObject *obj = mi_api_object_begin(mi_mem_strdup("geofeather"));

    mi_api_object_callback(geofeather_cb, (void *)paras);
    obj->visible =
    obj->reflection  =
	obj->refraction  =
  	obj->shadowmap = miTRUE;
	obj->shadow  = *mi_eval_integer(&paras->shadow_mode);
    obj->bbox_min.x = -0.5f;
    obj->bbox_min.y =  0.0f;
    obj->bbox_min.z = -0.1f;
    obj->bbox_max.x =  0.5f;
    obj->bbox_max.y =  2.0f;
    obj->bbox_max.z =  0.1f;
    tag = mi_api_object_end();
    mi_geoshader_add_result(result, tag);
    obj = (miObject *)mi_scene_edit(tag);
    obj->geo.placeholder_list.type = miOBJECT_HAIR;
    mi_scene_edit_end(tag);

    return miTRUE;
}

/*
 * geofeather callback. expands placeholder to a full feather object when the
 * bounding box of the instance is intersected by a ray, or scanline
 * rendered.
 */

DLLEXPORT miBoolean geofeather_cb(miTag tag, void *args)
{
    const char      *name;
    miObject        *obj;
    miHair_list     *hair;
    miScalar        *sarray;
    miGeoIndex      *harray;
    geofeather_t *paras = (geofeather_t *)args;
    int i = 0;
    int h;
//   int hcount = 41;
    /* when using the nbarbs parameter, instead of default here of 41 ... */
	int hcount = paras->hair_number;

    name = mi_api_tag_lookup(tag);
    mi_api_incremental(miTRUE);
    obj = mi_api_object_begin(mi_mem_strdup(name));
    obj->visible = obj->reflection = obj->refraction = obj->shadow = miTRUE;

    hair = mi_api_hair_begin();
 //  hair->approx = 5;				//need edit here for mr 3.7
    hair->degree = 3;               /* cubic                    */

    mi_api_hair_info(1, 'r', 1);    /* per-hair radius          */
    mi_api_hair_info(1, 't', 4);    /* per-vertex textures (4)  */
    mi_api_hair_info(1, 'm', 3);    /* per-vertex motion vector */

    /* here we add the geometry data. in a real program, this would be a
       very clever part. i just use numbers from a sample mi file i have. */
    sarray = mi_api_hair_scalars_begin(44*(hcount*2+1));
    i = 0;
    /* hair 1 is spine of feather */
    /* per-hair data for hair 1, here just the radius. */
    /* hair 1 control points 1-4, incuding per-vertex data. */
    sarray[i++] =  0.0f; sarray[i++] =  0.0f; sarray[i++] =  0.0f; /* vertex  */
	sarray[i++] =  0.01f;
    sarray[i++] =  1.0f; sarray[i++] =  1.0f; sarray[i++] =  1.0f; sarray[i++] =  1.0f; /* texture */
    sarray[i++] =  0.0f; sarray[i++] =  0.0f; sarray[i++] =  0.0f; /* motion  */

	
    sarray[i++] =  0.0f; sarray[i++] =  0.33f; sarray[i++] =  0.0f; /* vertex  */
	sarray[i++] =  0.02f;
    sarray[i++] =  1.0f; sarray[i++] =  1.0f; sarray[i++] =  1.0f; sarray[i++] =  0.9f; /* texture */
    sarray[i++] =  0.0f; sarray[i++] =  0.33f; sarray[i++] =  0.0f; /* motion  */

	
    sarray[i++] =  0.0f; sarray[i++] =  0.67f; sarray[i++] =  0.0f; /* vertex  */
	sarray[i++] =  0.01f;
    sarray[i++] =  1.0f; sarray[i++] =  1.0f; sarray[i++] =  1.0f; sarray[i++] =  0.8f; /* texture */
    sarray[i++] =  0.0f; sarray[i++] =  0.67f; sarray[i++] =  0.0f; /* motion  */

	 sarray[i++] =  0.0f; sarray[i++] =  1.0f; sarray[i++] =  0.0f; /* vertex  */
	sarray[i++] =  0.005f;
    sarray[i++] =  1.0f; sarray[i++] =  1.0f; sarray[i++] =  1.0f; sarray[i++] =  0.7f; /* texture */
    sarray[i++] =  0.0f; sarray[i++] =  1.0f; sarray[i++] =  0.0f; /* motion  */

    /* iterate for hairs along spine */

    for (h=0; h < hcount; h++) 
	{
      /* per-hair data for hair 2, ie. radius. */
      
      /* hair 2, 3, 4, ... control points 1-4, incuding per-vertex data. */
      sarray[i++] =  0.01f; sarray[i++] =  0.2f + h * 0.025f; sarray[i++] =  0.0f; /* vertex  */
	  sarray[i++] =  0.003f;
      sarray[i++] =  0.0f;  sarray[i++] =  0.6f;            sarray[i++] =  0.8f; sarray[i++] =  0.4f; /* texture */
      sarray[i++] =  0.01f; sarray[i++] =  0.2f + h * 0.025f; sarray[i++] =  0.0f; /* motion  */

		

      sarray[i++] =  0.11f; sarray[i++] =  0.3f + h * 0.025f; sarray[i++] =  0.0f; /* vertex  */
	  sarray[i++] =  0.01f;
      sarray[i++] =  0.0f;  sarray[i++] =  0.6f;            sarray[i++] =  0.8f; sarray[i++] =  0.34f; /* texture */
      sarray[i++] =  0.11f; sarray[i++] =  0.3f + h * 0.025f; sarray[i++] =  0.0f; /* motion  */

		

      sarray[i++] =  0.21f; sarray[i++] =  0.38f + h * 0.025f; sarray[i++] =  0.0f; /* vertex  */
	  sarray[i++] =  0.006f;
      sarray[i++] =  0.0f;  sarray[i++] =  0.6f;             sarray[i++] =  0.8f; sarray[i++] =  0.26f; /* texture */
      sarray[i++] =  0.21f; sarray[i++] =  0.38f + h * 0.025f; sarray[i++] =  0.0f; /* motion  */

		

      sarray[i++] =  0.31f; sarray[i++] =  0.46f + h * 0.025f; sarray[i++] =  0.0f; /* vertex  */
	  sarray[i++] =  0.001f;
      sarray[i++] =  0.0f;  sarray[i++] =  0.4f;             sarray[i++] =  0.8f; sarray[i++] =  0.0f; /* texture */
      sarray[i++] =  0.31f; sarray[i++] =  0.46f + h * 0.025f; sarray[i++] =  0.0f; /* motion  */

      /* mirrored per-hair data for hair 2, ie. radius. */
      
      /* mirror hair 2, 3, 4, ... control points 1-4, incuding per-vertex data. */
      sarray[i++] = -0.01f; sarray[i++] =  0.2f + h * 0.025f; sarray[i++] =  0.0f; /* vertex  */
	  sarray[i++] =  0.003f;
      sarray[i++] =  0.0f;  sarray[i++] =  0.6f;            sarray[i++] =  0.8f; sarray[i++] =  0.4f; /* texture */
      sarray[i++] = -0.01f; sarray[i++] =  0.2f + h * 0.025f; sarray[i++] =  0.0f; /* motion  */


      sarray[i++] = -0.11f; sarray[i++] =  0.3f + h * 0.025f; sarray[i++] =  0.0f; /* vertex  */
	  sarray[i++] =  0.01f;
      sarray[i++] =  0.0f;  sarray[i++] =  0.6f;            sarray[i++] =  0.8f; sarray[i++] =  0.34f; /* texture */
      sarray[i++] = -0.11f; sarray[i++] =  0.3f + h * 0.025f; sarray[i++] =  0.0f; /* motion  */

		
      sarray[i++] = -0.21f; sarray[i++] =  0.38f + h * 0.025f; sarray[i++] =  0.0f; /* vertex  */
	  sarray[i++] =  0.006f;
      sarray[i++] =  0.0f;  sarray[i++] =  0.6f;             sarray[i++] =  0.8f; sarray[i++] =  0.26f; /* texture */
      sarray[i++] = -0.21f; sarray[i++] =  0.38f + h * 0.025f; sarray[i++] =  0.0f; /* motion  */

		
      sarray[i++] = -0.31f; sarray[i++] =  0.46f + h * 0.025f; sarray[i++] =  0.0f; /* vertex  */
	  sarray[i++] =  0.001f;
      sarray[i++] =  0.0f;  sarray[i++] =  0.4f;             sarray[i++] =  0.8f; sarray[i++] =  0.20f; /* texture */
      sarray[i++] = -0.31f; sarray[i++] =  0.46f + h * 0.025f; sarray[i++] =  0.0f; /* motion  */
    }
    mi_api_hair_scalars_end( 44 * (hcount*2 + 1) );

    /* there are just the two hairs, ie. three numbers describe the sizes. */
    harray = mi_api_hair_hairs_begin(hcount*2 + 1 + 1);
    for (h=0; h < hcount*2 + 1 + 1; h++) {
      harray[h] = h * 44;
    }
    mi_api_hair_hairs_end();

    mi_api_hair_end();
    mi_api_object_end();

    return miTRUE;
}
}
