//  Copyright (c) 2008, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <math.h>
#include "shader.h"
#include "mi_shader_if.h"
#include "p_utils.h"

#ifdef DEBUG
#define miASSERT assert
#else
#define miASSERT(x)
#endif

typedef struct
{
	miBoolean		diffuse_en;
		miScalar		diffuse_coefficient;
		miScalar		back_diffuse_coefficient;
		miColor			diffuse_root;
		miColor			diffuse_tip;
		miScalar		saturation_variation;
		miScalar		color_variation;
		miScalar		gain_variation;
		miScalar		normal_blend_diffuse;

	miBoolean		specular_en;
		miScalar		specular_coefficient;
		miColor			prim_specular;
		miScalar		prim_shininess;
		miScalar		prim_deviation;

		miColor			sec_specular;
		miScalar		sec_shininess;
		miScalar		sec_deviation;

		miColor			tran_specular;
		miScalar		tran_shininess;
		miScalar		tran_deviation;

		miScalar		ior;
		miScalar		fresnel_amt;
		miScalar		normal_blend_specular;
		miScalar		directionality;

	miBoolean		trans_en;
		miColor			transparency_color;
		miScalar		trans_root;
		miScalar		trans_tip;
		miScalar		trans_cross;
		miScalar		trans_range;
		miScalar		trans_border;
		miInteger		trans_multiple;
		miScalar		trans_threshold;
		miInteger		trans_depth;

	miBoolean	use_indirect_illum;
		miColor		cInd;
		miScalar	saturation_indirect;
		int			visible_for_fg;

	miBoolean	use_ambient;
		miScalar	cAmb;
		miColor		ambient_color;

	miBoolean	use_occlusion;
		miScalar	effect_diffuse;
		miScalar	effect_specular;
		miScalar	effect_indirect;
		miScalar	effect_ambient;
		miColor		custom_occlusion;

	miBoolean	use_incandescence;
		miScalar	cInc;
		miColor		incandescence;

	miBoolean	use_SpecialC1;
		miColor		specialC1;
	miBoolean	use_SpecialC2;
		miColor		specialC2;
	miBoolean	use_SpecialC3;
		miColor		specialC3;
	miBoolean	use_SpecialC4;
		miColor		specialC4;
	miBoolean	use_SpecialC5;
		miColor		specialC5;
	miBoolean	use_SpecialC6;
		miColor		specialC6;
	miBoolean	use_SpecialC7;
		miColor		specialC7;
	miBoolean	use_SpecialC8;
		miColor		specialC8;
	miBoolean	use_SpecialC9;
		miColor		specialC9;
	miBoolean	use_SpecialC10;
		miColor		specialC10;

	miBoolean		use_z;
		miScalar 	near;
		miScalar 	far;
		miBoolean 	opacity_z;
		miBoolean 	use_float_z;
		miBoolean 	invert_z;

	miBoolean shading_pass;
	miBoolean color_occlusion_pass;
	miBoolean diffuse_pass;
	miBoolean diffuse_shadow_pass;
	miBoolean specular_pass;
	miBoolean specular_shadow_pass;
	miBoolean occlusion_pass;
	miBoolean indirect_illum_pass;
	miBoolean ambient_pass;
	miBoolean amb_occlusion_pass;
	miBoolean incandescence_pass;
	miBoolean z_pass;
//	miBoolean bent_normal_pass;	
	miBoolean specialC1_pass;
	miBoolean specialC2_pass;
	miBoolean specialC3_pass;
	miBoolean specialC4_pass;
	miBoolean specialC5_pass;
	miBoolean specialC6_pass;
	miBoolean specialC7_pass;
	miBoolean specialC8_pass;
	miBoolean specialC9_pass;
	miBoolean specialC10_pass;

	miBoolean fixShaveUV;

	int		mode;			// light mode: 0..2
	int		i_light;		// index of first light
	int		n_light;		// number of lights
	miTag		light[1];	// list of lights
}p_HairTK_t;

typedef struct		//struct that p_LightTK return
{
	miColor l_color_shd;		//light color with shadow
	miColor	l_color;			//light color
	miBoolean emit_diffuse;
	miBoolean emit_specular;
	char category[32];
}light_color;

typedef struct
{
	miScalar	sum_transparency;
	miInteger	level;
} p_hair_info;

// lookup tables
miScalar	*cosTable, *acosTable;

#define LUT_SIZE	0x3FF //1023

typedef struct		//struct for all output channels
{
	miColor	out_result;
	miColor	out_shading;
	miColor	out_color;
	miColor	out_diff_occlusion;
	miColor	out_diffuse;
	miColor	out_diffuse_shadow;
	miColor	out_specular;
	miColor	out_specular_shadow;
	miColor	out_indirect_illum;
	miColor	out_ambient;
	miColor	out_amb_occlusion;
	miColor	out_occlusion;
	miColor out_incandescence;
	miColor out_z;
//	miColor out_bent_normal;
	miColor out_specialC1;
	miColor out_specialC2;
	miColor out_specialC3;
	miColor out_specialC4;
	miColor out_specialC5;
	miColor out_specialC6;
	miColor out_specialC7;
	miColor out_specialC8;
	miColor out_specialC9;
	miColor out_specialC10;
}special_str_hair;


struct static_hair_param
{
	miBoolean	use_indirect_illum;
	int			visible_for_fg;
	miBoolean	use_ambient;
	miBoolean	use_occlusion;
	miBoolean	use_incandescence;
	miBoolean	use_z;
	miBoolean	invert_z;
	miBoolean	use_float_z;

	miBoolean	use_SpecialC1;	
	miBoolean	use_SpecialC2;
	miBoolean	use_SpecialC3;
	miBoolean	use_SpecialC4;
	miBoolean	use_SpecialC5;
	miBoolean	use_SpecialC6;
	miBoolean	use_SpecialC7;	
	miBoolean	use_SpecialC8;
	miBoolean	use_SpecialC9;
	miBoolean	use_SpecialC10;

	miBoolean shading_pass;
	int shading_buf;
	miBoolean color_pass;
	int color_buf;
	miBoolean diffuse_pass;
	int diffuse_buf;
	miBoolean diffuse_shadow_pass;
	int diffuse_shadow_buf;
	miBoolean specular_pass;
	int specular_buf;
	miBoolean specular_shadow_pass;
	int specular_shadow_buf;
	miBoolean diff_occlusion_pass;
	int diff_occlusion_buf;
	miBoolean indirect_illum_pass;
	int indirect_illum_buf;
	miBoolean ambient_pass;
	int ambient_buf;
	miBoolean amb_occlusion_pass;
	int amb_occlusion_buf;
	miBoolean incandescence_pass;
	int incandescence_buf;
	miBoolean z_pass;
	int z_buf;
	miBoolean specialC1_pass;
	int specialC1_buf;
	miBoolean specialC2_pass;
	int specialC2_buf;
	miBoolean specialC3_pass;
	int specialC3_buf;
	miBoolean specialC4_pass;
	int specialC4_buf;
	miBoolean specialC5_pass;
	int specialC5_buf;
	miBoolean specialC6_pass;
	int specialC6_buf;
	miBoolean specialC7_pass;
	int specialC7_buf;
	miBoolean specialC8_pass;
	int specialC8_buf;
	miBoolean specialC9_pass;
	int specialC9_buf;
	miBoolean specialC10_pass;
	int specialC10_buf;

	miBoolean fixShaveUV;
};

extern "C"
{
DLLEXPORT int p_HairTK_version(void){return 1;}

DLLEXPORT miBoolean p_HairTK_init(miState *state, p_HairTK_t *paras, miBoolean *inst_req)
{
	struct static_hair_param *sp, **spp;
	if (!paras)
	{
		*inst_req = miTRUE;

		// cosinus lookup table
		// possible values range from -pi/2 to +pi/2
		cosTable = (miScalar*)mi_mem_allocate(sizeof(miScalar) * LUT_SIZE);

		// arcosinus lookup table
		// possible values range from 0 to 1
		acosTable = (miScalar*)mi_mem_allocate(sizeof(miScalar) * LUT_SIZE);

		int i;
		for (i = 0; i < LUT_SIZE; i++)
		{
			cosTable[i] = (float)cos(((float)i / (float)LUT_SIZE) * M_PI - M_PI/2.0f);
			acosTable[i] = (float)acos((float)i / (float)LUT_SIZE);
		}

		return(miTRUE);
	}

	sp = (struct static_hair_param *)
	mi_mem_allocate(sizeof(struct static_hair_param));

	mi_query(miQ_FUNC_USERPTR, state, 0, &spp);
	*spp = sp;

	sp->use_indirect_illum = *mi_eval_boolean(&paras->use_indirect_illum);
	sp->visible_for_fg = *mi_eval_integer(&paras->visible_for_fg);
	sp->use_ambient = *mi_eval_boolean(&paras->use_ambient);
	sp->use_occlusion = *mi_eval_boolean(&paras->use_occlusion);
	sp->use_incandescence = *mi_eval_boolean(&paras->use_incandescence);
	sp->use_z = *mi_eval_boolean(&paras->use_z);
	sp->invert_z = *mi_eval_boolean(&paras->invert_z);
	sp->use_float_z = *mi_eval_boolean(&paras->use_float_z);
	sp->use_SpecialC1 = *mi_eval_boolean(&paras->use_SpecialC1);
	sp->use_SpecialC2 = *mi_eval_boolean(&paras->use_SpecialC2);
	sp->use_SpecialC3 = *mi_eval_boolean(&paras->use_SpecialC3);
	sp->use_SpecialC4 = *mi_eval_boolean(&paras->use_SpecialC4);
	sp->use_SpecialC5 = *mi_eval_boolean(&paras->use_SpecialC5);
	sp->use_SpecialC6 = *mi_eval_boolean(&paras->use_SpecialC6);
	sp->use_SpecialC7 = *mi_eval_boolean(&paras->use_SpecialC7);
	sp->use_SpecialC8 = *mi_eval_boolean(&paras->use_SpecialC8);
	sp->use_SpecialC9 = *mi_eval_boolean(&paras->use_SpecialC9);
	sp->use_SpecialC10 = *mi_eval_boolean(&paras->use_SpecialC10);

	
	size_t index = -1;
	sp->shading_pass = *mi_eval_boolean(&paras->shading_pass);
	sp->shading_buf = (int)index;
	sp->color_pass = *mi_eval_boolean(&paras->color_occlusion_pass);
	sp->color_buf = (int)index;
	sp->diffuse_pass = *mi_eval_boolean(&paras->diffuse_pass);
	sp->diffuse_buf = (int)index;	
	sp->diffuse_shadow_pass = *mi_eval_boolean(&paras->diffuse_shadow_pass);
	sp->diffuse_shadow_buf  = (int)index;	
	sp->specular_pass = *mi_eval_boolean(&paras->specular_pass);
	sp->specular_buf = (int)index;
	sp->specular_shadow_pass = *mi_eval_boolean(&paras->specular_shadow_pass);
	sp->specular_shadow_buf = (int)index;
	sp->diff_occlusion_pass = *mi_eval_boolean(&paras->occlusion_pass);
	sp->diff_occlusion_buf = (int)index;
	sp->indirect_illum_pass = *mi_eval_boolean(&paras->indirect_illum_pass);
	sp->indirect_illum_buf = (int)index;
	sp->ambient_pass = *mi_eval_boolean(&paras->ambient_pass);
	sp->ambient_buf = (int)index;
	sp->amb_occlusion_pass = *mi_eval_boolean(&paras->amb_occlusion_pass);
	sp->amb_occlusion_buf = (int)index;
//	sp->reflect_pass = *mi_eval_boolean(&paras->reflect_pass);
//	sp->reflect_buf = (int)index;
//	sp->refract_pass = *mi_eval_boolean(&paras->refract_pass);
//	sp->refract_buf = (int)index;
	sp->incandescence_pass = *mi_eval_boolean(&paras->incandescence_pass);
	sp->incandescence_buf = (int)index;
	sp->z_pass = *mi_eval_boolean(&paras->z_pass);
	sp->z_buf = (int)index;
//	sp->bent_normal_pass = *mi_eval_boolean(&paras->bent_normal_pass);
//	sp->bent_normal_buf = (int)index;
	sp->specialC1_pass = *mi_eval_boolean(&paras->specialC1_pass);
	sp->specialC1_buf = (int)index;
	sp->specialC2_pass = *mi_eval_boolean(&paras->specialC2_pass);
	sp->specialC2_buf = (int)index;
	sp->specialC3_pass = *mi_eval_boolean(&paras->specialC3_pass);
	sp->specialC3_buf = (int)index;
	sp->specialC4_pass = *mi_eval_boolean(&paras->specialC4_pass);
	sp->specialC4_buf = (int)index;
	sp->specialC5_pass = *mi_eval_boolean(&paras->specialC5_pass);
	sp->specialC5_buf = (int)index;
	sp->specialC6_pass = *mi_eval_boolean(&paras->specialC6_pass);
	sp->specialC6_buf = (int)index;
	sp->specialC7_pass = *mi_eval_boolean(&paras->specialC7_pass);
	sp->specialC7_buf = (int)index;
	sp->specialC8_pass = *mi_eval_boolean(&paras->specialC8_pass);
	sp->specialC8_buf = (int)index;
	sp->specialC9_pass = *mi_eval_boolean(&paras->specialC9_pass);
	sp->specialC9_buf = (int)index;
	sp->specialC10_pass = *mi_eval_boolean(&paras->specialC10_pass);
	sp->specialC10_buf = (int)index;

	/*
	mi::shader::Access_fb fbc(state->camera->buffertag);
	size_t index;

	sp->shading_pass = *mi_eval_boolean(&paras->shading_pass);
	index = -1;
	fbc->get_index("tk_shading", index); 
	sp->shading_buf = (int)index;

	sp->color_pass = *mi_eval_boolean(&paras->color_occlusion_pass);
	index = -1;
	fbc->get_index("tk_color", index); 
	sp->color_buf = (int)index;

	sp->diffuse_pass = *mi_eval_boolean(&paras->diffuse_pass);
	index = -1;
	fbc->get_index("tk_diffuse", index); 
	sp->diffuse_buf = (int)index;
	
	sp->diffuse_shadow_pass = *mi_eval_boolean(&paras->diffuse_shadow_pass);
	index = -1;
	fbc->get_index("tk_shadow", index); 
	sp->diffuse_shadow_buf  = (int)index;
	
	sp->specular_pass = *mi_eval_boolean(&paras->specular_pass);
	index = -1;
	fbc->get_index("tk_specular", index); 
	sp->specular_buf = (int)index;

	sp->specular_shadow_pass = *mi_eval_boolean(&paras->specular_shadow_pass);
	index = -1;
	fbc->get_index("tk_spec_shd", index); 
	sp->specular_shadow_buf = (int)index;

	sp->diff_occlusion_pass = *mi_eval_boolean(&paras->occlusion_pass);
	index = -1;
	fbc->get_index("tk_diff_occl", index); 
	sp->diff_occlusion_buf = (int)index;

	sp->indirect_illum_pass = *mi_eval_boolean(&paras->indirect_illum_pass);
	index = -1;
	fbc->get_index("tk_indirect", index); 
	sp->indirect_illum_buf = (int)index;

	sp->ambient_pass = *mi_eval_boolean(&paras->ambient_pass);
	index = -1;
	fbc->get_index("tk_ambient", index); 
	sp->ambient_buf = (int)index;

	sp->amb_occlusion_pass = *mi_eval_boolean(&paras->amb_occlusion_pass);
	index = -1;
	fbc->get_index("tk_amb_occlus", index); 
	sp->amb_occlusion_buf = (int)index;

	sp->incandescence_pass = *mi_eval_boolean(&paras->incandescence_pass);
	index = -1;
	fbc->get_index("tk_incand", index); 
	sp->incandescence_buf = (int)index;

	sp->z_pass = *mi_eval_boolean(&paras->z_pass);
	index = -1;
	fbc->get_index("tk_z", index); 
	sp->z_buf = (int)index;

//	sp->bent_normal_pass = *mi_eval_boolean(&paras->bent_normal_pass);
//	index = -1;
//	fbc->get_index("tk_bent", index); 
//	sp->bent_normal_buf = (int)index;

	sp->specialC1_pass = *mi_eval_boolean(&paras->specialC1_pass);
	index = -1;
	fbc->get_index("tk_specialC1", index); 
	sp->specialC1_buf = (int)index;

	sp->specialC2_pass = *mi_eval_boolean(&paras->specialC2_pass);
	index = -1;
	fbc->get_index("tk_specialC2", index); 
	sp->specialC2_buf = (int)index;

	sp->specialC3_pass = *mi_eval_boolean(&paras->specialC3_pass);
	index = -1;
	fbc->get_index("tk_specialC3", index); 
	sp->specialC3_buf = (int)index;

	sp->specialC4_pass = *mi_eval_boolean(&paras->specialC4_pass);
	index = -1;
	fbc->get_index("tk_specialC4", index); 
	sp->specialC4_buf = (int)index;

	sp->specialC5_pass = *mi_eval_boolean(&paras->specialC5_pass);
	index = -1;
	fbc->get_index("tk_specialC5", index); 
	sp->specialC5_buf = (int)index;

	sp->specialC6_pass = *mi_eval_boolean(&paras->specialC6_pass);
	index = -1;
	fbc->get_index("tk_specialC6", index); 
	sp->specialC6_buf = (int)index;

	sp->specialC7_pass = *mi_eval_boolean(&paras->specialC7_pass);
	index = -1;
	fbc->get_index("tk_specialC7", index); 
	sp->specialC7_buf = (int)index;

	sp->specialC8_pass = *mi_eval_boolean(&paras->specialC8_pass);
	index = -1;
	fbc->get_index("tk_specialC8", index); 
	sp->specialC8_buf = (int)index;

	sp->specialC9_pass = *mi_eval_boolean(&paras->specialC9_pass);
	index = -1;
	fbc->get_index("tk_specialC9", index); 
	sp->specialC9_buf = (int)index;

	sp->specialC10_pass = *mi_eval_boolean(&paras->specialC10_pass);
	index = -1;
	fbc->get_index("tk_specialC10", index); 
	sp->specialC10_buf = (int)index;
*/
	sp->fixShaveUV = *mi_eval_boolean(&paras->fixShaveUV);

	return miTRUE;
}

DLLEXPORT miBoolean p_HairTK(miColor *result, miState *state, p_HairTK_t *paras)
{
	miBoolean	trans_en;
	miColor		trans_col;
	miScalar	trans_root, trans_tip, trans_cross, trans_range, trans_border;
	miScalar	trans_threshold;
	miInteger	trans_depth;

	miScalar	trans = 0.0f;

	miScalar	ior;
	miScalar	fresnel_amt;
	miScalar	directionality;

	miColor		diffuse_root;	
	miColor		*diffuse_tip;	
	miColor		prim_specular, tran_specular, sec_specular;
	miScalar	prim_shininess, tran_shininess, sec_shininess;	
	miScalar	prim_deviation, tran_deviation, sec_deviation;

	miVector	v;
	miColor		sum_light;		// summed sample colors
	miColor		sum_light_shd;		// summed sample colors
	miColor		sum_specular;		// summed specular colors
	miColor		sum_specular_shd;		// summed specular colors
	miScalar	diffuse_dot;
	miScalar	light_dot;
	miScalar	view_dot;
	miScalar	back_dot;
	miVector	light_dir;
	miInteger	 samples;		// for light sampling
	miInteger	 mode;
	miInteger	 i_light;
	miInteger	 n_light;
	miTag*		 light;

	light_color	light_col;		//struct for light shader return

	light_color *lightTK_pt = NULL;
	int size_light_color = 0;	

    // illegal call, return to prevent crash
	if(state->type == miRAY_DISPLACE || state->type == miRAY_SHADOW)
		return miFALSE;

	struct static_hair_param *sp, **spp;
	miASSERT(state->shader);
	mi_query(miQ_FUNC_USERPTR, state, 0, &spp);
	sp = *spp;

	if(sp->fixShaveUV)	//restore original surface UV in state->tex_list[0] that was overrided by Shave
	{
		state->tex_list[0].x = state->tex_list[2].y;
		state->tex_list[0].y = state->tex_list[2].z;
	}

	// first of all, grab the transparency settings
	trans_en	= *mi_eval_boolean(&paras->trans_en);
	trans_threshold = *mi_eval_scalar(&paras->trans_threshold);
	trans_depth = *mi_eval_integer(&paras->trans_depth);


	if(state->type == miRAY_FINALGATHER)
	{
		if(sp->visible_for_fg == 1)				//opaque hair
			trans_en = miFALSE;
		else if (sp->visible_for_fg == 2)		//invisible hair (possible some bug here)
			return (mi_trace_continue(result, state));
	}


	special_str_hair special;		//struct for output special channel


	special.out_shading.r = special.out_shading.g = special.out_shading.b = 0;
	special.out_indirect_illum.r = special.out_indirect_illum.g = special.out_indirect_illum.b = 0;
	special.out_color.r = special.out_color.g = special.out_color.b = 0;
	special.out_ambient.r = special.out_ambient.g = special.out_ambient.b = 0;
	special.out_amb_occlusion.r = special.out_amb_occlusion.g = special.out_amb_occlusion.b = 0;
	special.out_occlusion.r = special.out_occlusion.g = special.out_occlusion.b = 0;
	special.out_diffuse_shadow.r = special.out_diffuse_shadow.g = special.out_diffuse_shadow.b = 0;
	special.out_diffuse.r = special.out_diffuse.g = special.out_diffuse.b = 0;
	special.out_specular.r = special.out_specular.g = special.out_specular.b = 0;
	special.out_specular_shadow.r = special.out_specular_shadow.g = special.out_specular_shadow.b = 0;
//	special.out_bent_normal.r = special.out_bent_normal.g =  special.out_bent_normal.b = 0;
	special.out_z.r = special.out_z.g = special.out_z.b = special.out_z.a = 0;
	special.out_diff_occlusion.r = special.out_diff_occlusion.g = special.out_diff_occlusion.b = 1;
	special.out_incandescence.r = special.out_incandescence.g = special.out_incandescence.b = 0;
	special.out_specialC1.r = special.out_specialC1.g = special.out_specialC1.b = 0;
	special.out_specialC2.r = special.out_specialC2.g = special.out_specialC2.b = 0;
	special.out_specialC3.r = special.out_specialC3.g = special.out_specialC3.b = 0;
	special.out_specialC4.r = special.out_specialC4.g = special.out_specialC4.b = 0;
	special.out_specialC5.r = special.out_specialC5.g = special.out_specialC5.b = 0;
	special.out_specialC6.r = special.out_specialC6.g = special.out_specialC6.b = 0;
	special.out_specialC7.r = special.out_specialC7.g = special.out_specialC7.b = 0;
	special.out_specialC8.r = special.out_specialC8.g = special.out_specialC8.b = 0;
	special.out_specialC9.r = special.out_specialC9.g = special.out_specialC9.b = 0;
	special.out_specialC10.r = special.out_specialC10.g = special.out_specialC10.b = 0;

	result->r = result->g = result->b = result->a = 0;
	trans_col.r = trans_col.g = trans_col.b = trans_col.a = 0;

//	miRc_intersection *priSave = state->pri;
//	miGeoScalar	offset;

	// check if we have a shader state set
	p_hair_info *mh = NULL;
	p_hair_info mh_data;
	bool mh_new = true;

	if(state->options->scanline != 'r')
	{
		mh = (p_hair_info *)mi_shaderstate_get(state, "p_hair_stat", NULL);
		if (mh != NULL)
		{
			mh_data = *mh;	
			mh_new = false;
//			if(mh_data.sum_transparency < trans_threshold || (mh_data.level > trans_depth && trans_depth))
//				trans_en = miFALSE;
		}
		else  //mh not created, just create it
		{
//			mh_new = true;
			mh_data.sum_transparency = 1.0f;
			mh_data.level = 0;
		}
	}
	else	//resterizer
	{
//		mh_new = true;
		mh_data.sum_transparency = 1.0f;
		mh_data.level = 0;
	}

	// if so, and one of our break conditions is true
//	if (mh && (mh_data.sum_transparency < trans_threshold || (mh_data.level > trans_depth && trans_depth > 0)))
//		trans_en = miFALSE;

	// compute our multiplicity factor, and the crosslocation / costheta
//	trans_multiple = *mi_eval_integer(&paras->trans_multiple);

	float hair_width = 0;
	float hair_length = 0;
	if(state->bary[0] == state->bary[0])
		hair_width = state->bary[0];
	else
		mi_warning("p_HairTK: NaN or infinite value in state->bary[0]. Replaced with zero");

	if(state->bary[1] == state->bary[1])
		hair_length = state->bary[1];
	else
		mi_warning("p_HairTK: NaN or infinite value in state->bary[1]. Replaced with zero");


	float crossloc = (float)fmod((double)(hair_width * *mi_eval_integer(&paras->trans_multiple)), 1.0);
//	float costheta = (float)cos(crossloc * M_PI - M_PI_2);
	float costheta	= cosTable[(int)(crossloc * LUT_SIZE)];

	float inv_alpha = 0;
	if (trans_en)
	{
		// evaluate all the transparency options
		trans_col	= *mi_eval_color(&paras->transparency_color);
		trans_root	= *mi_eval_scalar(&paras->trans_root);
		trans_tip	= *mi_eval_scalar(&paras->trans_tip);

		trans_range	= *mi_eval_scalar(&paras->trans_range);
		trans_cross	= *mi_eval_scalar(&paras->trans_cross);
		trans_border = *mi_eval_scalar(&paras->trans_border);

		float interp = p_smooth_step(p_clamp_0_1(trans_cross - trans_range), p_clamp_0_1(trans_cross + trans_range), hair_length);

		trans = trans_root * (1.0f - interp) + trans_tip * interp;

		if (trans_border > miAlphaEPS)
			trans += (1.0f - costheta) * trans_border * (1.0f - trans);

//		trans_col.r = p_max(trans_col.r, trans);
//		trans_col.g = p_max(trans_col.g, trans);
//		trans_col.b = p_max(trans_col.b, trans);
//		trans_col.a = p_max(trans_col.a, trans);

//		inv_alpha = p_color_maxcomp(&trans_col);

		inv_alpha = (trans > trans_col.r) ? trans : trans_col.r;
		inv_alpha = (inv_alpha > trans_col.g) ? inv_alpha : trans_col.g;
		inv_alpha = (inv_alpha > trans_col.b) ? inv_alpha : trans_col.b;

		if(inv_alpha <= miAlphaEPS)
			trans_en = miFALSE;
		
		if(trans_en && state->options->scanline != 'r')
		{
			// now that we have our transparency, compute the contribution to the final pixel
			mh_data.sum_transparency *= inv_alpha;
			mh_data.level++;
			if(mh_data.sum_transparency < trans_threshold || mh_data.level > trans_depth && trans_depth)
			{
				inv_alpha = 0.0f;
				trans_en = miFALSE;
			}
//			mi_shaderstate_set(state, "p_hair_stat", &mh_data, sizeof(mh_data), miSS_LIFETIME_EYERAY);
		}
	}


//	float inv_alpha = (trans_col.r + trans_col.g + trans_col.b)*0.33333f;
//	inv_alpha = p_color_maxcomp(&trans_col);
	float alpha = 1 -inv_alpha;
	result->a = alpha;

//	if (!trans_en || inv_alpha < 1.0f -trans_threshold)
	if (!trans_en || alpha > miAlphaEPS)  // hair is not comletely trasparent, so we compute shading
	{
		float	fresnel;
		float	smooth_interp = 0;
		float	diffuse_coef = 0;
		float	back_diffuse_coef = 0;
		float	specular_coef = 0;
		float	dim_fresnel = 0;
		float	diff_blend = 0;
		float	spec_blend = 0;

		miBoolean diffuse_en = *mi_eval_boolean(&paras->diffuse_en);
		miBoolean specular_en = *mi_eval_boolean(&paras->specular_en);

		mode       = *mi_eval_integer(&paras->mode);
		i_light    = *mi_eval_integer(&paras->i_light);
		n_light    = *mi_eval_integer(&paras->n_light);
		light      =  mi_eval_tag(paras->light) + i_light;

		// correct light list, if requested
		if (mode == 1)
			mi_inclusive_lightlist(&n_light, &light, state);
		else if (mode == 2)
			mi_exclusive_lightlist(&n_light, &light, state);
		else if (mode == 4) 
		{
			n_light = 0;
			light = 0;
		}

		if(diffuse_en)
		{
			diffuse_coef = *mi_eval_scalar(&paras->diffuse_coefficient);
			back_diffuse_coef = *mi_eval_scalar(&paras->back_diffuse_coefficient);
			if(diffuse_coef > miEPS)
			{
				diff_blend = *mi_eval_scalar(&paras->normal_blend_diffuse);
				diffuse_root = *mi_eval_color(&paras->diffuse_root);
				diffuse_tip = mi_eval_color(&paras->diffuse_tip);

				smooth_interp = p_smooth_step(0, 1, state->bary[1]);
				diffuse_root.r = diffuse_root.r * (1 -smooth_interp) + diffuse_tip->r * smooth_interp;
				diffuse_root.g = diffuse_root.g * (1 -smooth_interp) + diffuse_tip->g * smooth_interp;
				diffuse_root.b = diffuse_root.b * (1 -smooth_interp) + diffuse_tip->b * smooth_interp;


				miScalar saturation_variation = *mi_eval_scalar(&paras->saturation_variation);
				miScalar color_variation = *mi_eval_scalar(&paras->color_variation);
				miScalar gain_variation = *mi_eval_scalar(&paras->gain_variation);

				if(color_variation > miEPS || saturation_variation > miEPS || gain_variation > miEPS)
				{
					miUint index = 0;
					if(mi_query(miQ_PRI_INDEX, state, NULL, &index))
					{
						if(saturation_variation > miEPS)
						{
							float rand_value;
							p_srand_float(&rand_value, index);
							float grey = (float)(diffuse_root.r + diffuse_root.g + diffuse_root.b) * 0.333333f;
							rand_value = (rand_value * 2 -1) * saturation_variation + 1;

							diffuse_root.r = diffuse_root.r * rand_value + grey * (1-rand_value);
							diffuse_root.g = diffuse_root.g * rand_value + grey * (1-rand_value);
							diffuse_root.b = diffuse_root.b * rand_value + grey * (1-rand_value);
						}
						if(color_variation > miEPS || gain_variation > miEPS)
						{
							miColor rand_value;
							p_srand_color(&rand_value, index);
							diffuse_root.r += color_variation * (rand_value.r * 2 - 1);
							diffuse_root.g += color_variation * (rand_value.g * 2 - 1);
							diffuse_root.b += color_variation * (rand_value.b * 2 - 1);

							if(gain_variation > miEPS)
							{
								gain_variation = 1 + ((rand_value.a * 2 - 1) * gain_variation);
								diffuse_root.r *= gain_variation;
								diffuse_root.g *= gain_variation;
								diffuse_root.b *= gain_variation;
							}
						}
					}
				}

				special.out_color.r = diffuse_root.r ? diffuse_root.r : 0;
				special.out_color.g = diffuse_root.g ? diffuse_root.g : 0;
				special.out_color.b = diffuse_root.b ? diffuse_root.b : 0;
			}
			else
				diffuse_en = miFALSE;
		}

		if(specular_en)
		{
			specular_coef = *mi_eval_scalar(&paras->specular_coefficient);
			if(specular_coef > miEPS)
			{
				spec_blend = *mi_eval_scalar(&paras->normal_blend_specular);
				prim_specular	= *mi_eval_color(&paras->prim_specular);
				tran_specular	= *mi_eval_color(&paras->tran_specular);
				sec_specular	= *mi_eval_color(&paras->sec_specular);
				// THIS IS SCALED DOWN, was far too much otherwise
				prim_shininess	= *mi_eval_scalar(&paras->prim_shininess) * 0.125f;
				tran_shininess	= *mi_eval_scalar(&paras->tran_shininess) * 0.125f;
				sec_shininess	= *mi_eval_scalar(&paras->sec_shininess) * 0.125f;

				if(prim_shininess != 0)
					prim_shininess = 1.0f / prim_shininess;
				if(tran_shininess != 0)
					tran_shininess = 1.0f / tran_shininess;
				if(sec_shininess != 0)
					sec_shininess = 1.0f / sec_shininess;



				prim_deviation	= *mi_eval_scalar(&paras->prim_deviation);
				tran_deviation	= *mi_eval_scalar(&paras->tran_deviation);
				sec_deviation	= *mi_eval_scalar(&paras->sec_deviation);

				directionality	= *mi_eval_scalar(&paras->directionality);

				fresnel_amt	= *mi_eval_scalar(&paras->fresnel_amt);
				ior			= *mi_eval_scalar(&paras->ior);

				fresnel = p_fresnel(1.0f, ior, costheta);
				fresnel = p_clamp_0_1(fresnel);
				fresnel = 1.0f - fresnel_amt + fresnel * fresnel_amt;

				dim_fresnel = 1.0f - fresnel_amt + (1.0f - fresnel) * fresnel_amt;
			}
			else
				specular_en = miFALSE;
		}


		// save the original normal
		miVector normal = state->normal;

		// radiance
		// set state->normal to base normal for irradiance calc
		state->normal = state->normal_geom;
		
		// grab our view vector
		v = state->dir;
		// invert it
		mi_vector_neg(&v);

		// grab the hair tangent
		miVector tang, base_tang = state->derivs[0];
		// normalize it
		mi_vector_normalize(&base_tang);

		// and set the one in the state to 0,0,0, to kill any normal optimizations by mray
		// HAXXX, doesn't correctly work with detail shadows, therefore leave user option.
//		if (0)
//		{
//			offset = 0.01;
	//		offset = (miGeoScalar)*mi_eval_scalar(&paras->m_detail_bias);
//			priSave = state->pri;
//			state->pri = 0;
//			mi_ray_offset(state, &offset);
//			state->normal = normal;
//		}
//		else
			state->normal.x = state->normal.y = state->normal.z = 0;

		// and bend it for our normal blending purposes
		tang.x = base_tang.x * (1.0f - spec_blend) + normal.x * spec_blend;
		tang.y = base_tang.y * (1.0f - spec_blend) + normal.y * spec_blend;
		tang.z = base_tang.z * (1.0f - spec_blend) + normal.z * spec_blend;

		mi_vector_normalize(&tang);

		miVector	diff_norm;

		// also bend the normal used for our diffuse calcs
		diff_norm.x = normal.x * (1.0f - diff_blend) + state->normal_geom.x * diff_blend;
		diff_norm.y = normal.y * (1.0f - diff_blend) + state->normal_geom.y * diff_blend;
		diff_norm.z = normal.z * (1.0f - diff_blend) + state->normal_geom.z * diff_blend;

		mi_vector_normalize(&diff_norm);

		if (mode == 4 || n_light) 
		{
			for (mi::shader::LightIterator iter(state, light, n_light); !iter.at_end(); ++iter) 
			{
				sum_light_shd.r = sum_light_shd.g = sum_light_shd.b = 0;
				sum_light.r = sum_light.g = sum_light.b = 0;
				sum_specular.r = sum_specular.g = sum_specular.b = 0;
				sum_specular_shd.r = sum_specular_shd.g = sum_specular_shd.b = 0;

				while (iter->sample()) 
				{
					diffuse_dot = 0.0f;
					view_dot = 0.0f;
					back_dot = 0.0f;
					light_dot = 0.0f;
					iter->get_contribution(&light_col.l_color_shd);
					light_dir = iter->get_direction();

//					size_light_color = 0;
					lightTK_pt = (light_color *)mi_shaderstate_get(state, lightsTK_key, NULL);  
					if (lightTK_pt != NULL)
						light_col = *lightTK_pt;

					if(lightTK_pt != NULL) //Yes, it's LightTK! Good. We compute shading twice, with light only and with light+shadow
					{
						if(light_col.emit_diffuse && diffuse_en)
						{
	//						diffuse_dot = mi_vector_dot(&diff_norm, &light_dir) * 0.5f + 0.5f;
							diffuse_dot = mi_vector_dot(&diff_norm, &light_dir);
							if(diffuse_dot > 0)
								diffuse_dot = diffuse_dot * 0.7f + 0.3f;
							else   //back diffuse
								diffuse_dot = -diffuse_dot * back_diffuse_coef * 0.7f + 0.3f;

							sum_light_shd.r += light_col.l_color_shd.r * diffuse_dot;
							sum_light_shd.g += light_col.l_color_shd.g * diffuse_dot;
							sum_light_shd.b += light_col.l_color_shd.b * diffuse_dot;

							sum_light.r += light_col.l_color.r * diffuse_dot;
							sum_light.g += light_col.l_color.g * diffuse_dot;
							sum_light.b += light_col.l_color.b * diffuse_dot;

						}
						if(light_col.emit_specular && specular_en)
						{							
							miVector	h, norm;

							view_dot = mi_vector_dot(&v, &light_dir) * 0.7f + 0.3f;
							back_dot = 1.0f - view_dot;
							
							view_dot = view_dot * directionality + (1.0f - directionality);
							back_dot = back_dot * directionality + (1.0f - directionality);

							// longitudinal specularity is computed through the dot product between the
							// hair tangent tang, and the incident light direction light_dir
							// maximum highlights are achieved when view.tang == light.tang
							mi_vector_add(&h, &v, &light_dir);
							mi_vector_normalize(&h);

							// for normal blending, also compute a standard dot product between
							// the halfway vector and the base tangent of the hair

							// apply deviation by tilting the halfway vector more towards tang
							norm.x = h.x + base_tang.x * prim_deviation;
							norm.y = h.y + base_tang.y * prim_deviation;
							norm.z = h.z + base_tang.z * prim_deviation;

							mi_vector_normalize(&norm);
							light_dot = mi_vector_dot(&tang, &norm);
							light_dot = 1.0f - light_dot*light_dot;

							// compute PRIMARY specular highlight, with fresnel falloff
							//light_dot = p_gauss_dist_norm(1.0f / prim_shininess, (miScalar)acos(light_dot)) * fresnel;
							//light_dot = p_gauss_dist_norm(prim_shininess, (miScalar)acos(light_dot)) * fresnel;
							light_dot = p_gauss_dist_norm(prim_shininess, acosTable[(int)(light_dot * LUT_SIZE)]) * fresnel;

							// specular
							sum_specular_shd.r += light_col.l_color_shd.r * light_dot * prim_specular.r * view_dot;
							sum_specular_shd.g += light_col.l_color_shd.g * light_dot * prim_specular.g * view_dot;
							sum_specular_shd.b += light_col.l_color_shd.b * light_dot * prim_specular.b * view_dot;

							sum_specular.r += light_col.l_color.r * light_dot * prim_specular.r * view_dot;
							sum_specular.g += light_col.l_color.g * light_dot * prim_specular.g * view_dot;
							sum_specular.b += light_col.l_color.b * light_dot * prim_specular.b * view_dot;

							// compute SECONDARY specular highlight

							// apply deviation
							norm.x = h.x + base_tang.x * sec_deviation;
							norm.y = h.y + base_tang.y * sec_deviation;
							norm.z = h.z + base_tang.z * sec_deviation;

							mi_vector_normalize(&norm);

							light_dot = mi_vector_dot(&tang, &norm);
							light_dot = 1.0f - light_dot*light_dot;

				//			light_dot = p_gauss_dist_norm(sec_shininess, (miScalar)acos(light_dot));
				//			light_dot = p_gauss_dist_norm(1.0f / sec_shininess, acosTable[(int)(light_dot*LUT_SIZE)]);
							light_dot = p_gauss_dist_norm(sec_shininess, acosTable[(int)(light_dot*LUT_SIZE)]);
							

							sum_specular_shd.r += light_col.l_color_shd.r * light_dot * sec_specular.r * view_dot;
							sum_specular_shd.g += light_col.l_color_shd.g * light_dot * sec_specular.g * view_dot;
							sum_specular_shd.b += light_col.l_color_shd.b * light_dot * sec_specular.b * view_dot;

							sum_specular.r += light_col.l_color.r * light_dot * sec_specular.r * view_dot;
							sum_specular.g += light_col.l_color.g * light_dot * sec_specular.g * view_dot;
							sum_specular.b += light_col.l_color.b * light_dot * sec_specular.b * view_dot;


							// compute TRANSLUCENT specular highlight

							// recompute halfway vector, since we now base it on the
							// inverted view direction

							mi_vector_add(&h, &state->dir, &light_dir);
							mi_vector_normalize(&h);

							norm.x = h.x + base_tang.x * tran_deviation;
							norm.y = h.y + base_tang.y * tran_deviation;
							norm.z = h.z + base_tang.z * tran_deviation;

							mi_vector_normalize(&norm);
							light_dot = mi_vector_dot(&tang, &norm);
							light_dot = 1.0f - light_dot*light_dot;

					//		light_dot = p_gauss_dist_norm(tran_shininess, (miScalar)acos(light_dot)) * dim_fresnel;
							light_dot = p_gauss_dist_norm(tran_shininess, acosTable[(int)(light_dot*LUT_SIZE)]) * dim_fresnel;

							// specular translucent
							sum_specular_shd.r += light_col.l_color_shd.r * light_dot * tran_specular.r * back_dot;
							sum_specular_shd.g += light_col.l_color_shd.g * light_dot * tran_specular.g * back_dot;
							sum_specular_shd.b += light_col.l_color_shd.b * light_dot * tran_specular.b * back_dot;

							sum_specular.r += light_col.l_color.r * light_dot * tran_specular.r * back_dot;
							sum_specular.g += light_col.l_color.g * light_dot * tran_specular.g * back_dot;
							sum_specular.b += light_col.l_color.b * light_dot * tran_specular.b * back_dot;
						}
					}
					else //It's not LightTK! Bad, but should work. It's work like any other common shader.
					{
						if(diffuse_en)
						{
	//						diffuse_dot = mi_vector_dot(&diff_norm, &light_dir) * 0.5f + 0.5f;
							diffuse_dot = mi_vector_dot(&diff_norm, &light_dir);
							if(diffuse_dot > 0)
								diffuse_dot = diffuse_dot * 0.7f + 0.3f;
							else   //back diffuse
								diffuse_dot = -diffuse_dot * back_diffuse_coef * 0.7f + 0.3f;

							sum_light_shd.r += light_col.l_color_shd.r * diffuse_dot;
							sum_light_shd.g += light_col.l_color_shd.g * diffuse_dot;
							sum_light_shd.b += light_col.l_color_shd.b * diffuse_dot;
						}

						if(specular_en)
						{
							miVector	h, norm;

							view_dot = mi_vector_dot(&v, &light_dir) * 0.7f + 0.3f;
							back_dot = 1.0f - view_dot;
							
							view_dot = view_dot * directionality + (1.0f - directionality);
							back_dot = back_dot * directionality + (1.0f - directionality);

							// longitudinal specularity is computed through the dot product between the
							// hair tangent tang, and the incident light direction light_dir
							// maximum highlights are achieved when view.tang == light.tang
							mi_vector_add(&h, &v, &light_dir);
							mi_vector_normalize(&h);

							// for normal blending, also compute a standard dot product between
							// the halfway vector and the base tangent of the hair

							// apply deviation by tilting the halfway vector more towards tang
							norm.x = h.x + base_tang.x * prim_deviation;
							norm.y = h.y + base_tang.y * prim_deviation;
							norm.z = h.z + base_tang.z * prim_deviation;

							mi_vector_normalize(&norm);
							light_dot = mi_vector_dot(&tang, &norm);
							light_dot = 1.0f - light_dot*light_dot;

							// compute PRIMARY specular highlight, with fresnel falloff
							//light_dot = p_gauss_dist_norm(1.0f / prim_shininess, (miScalar)acos(light_dot)) * fresnel;
						//	light_dot = p_gauss_dist_norm(prim_shininess, (miScalar)acos(light_dot)) * fresnel;
							light_dot = p_gauss_dist_norm(prim_shininess, acosTable[(int)(light_dot * LUT_SIZE)]) * fresnel;


							// specular
							sum_specular_shd.r += light_col.l_color_shd.r * light_dot * prim_specular.r * view_dot;
							sum_specular_shd.g += light_col.l_color_shd.g * light_dot * prim_specular.g * view_dot;
							sum_specular_shd.b += light_col.l_color_shd.b * light_dot * prim_specular.b * view_dot;

							// compute SECONDARY specular highlight

							// apply deviation
							norm.x = h.x + base_tang.x * sec_deviation;
							norm.y = h.y + base_tang.y * sec_deviation;
							norm.z = h.z + base_tang.z * sec_deviation;

							mi_vector_normalize(&norm);

							light_dot = mi_vector_dot(&tang, &norm);
							light_dot = 1.0f - light_dot*light_dot;

					//		light_dot = p_gauss_dist_norm(sec_shininess, (miScalar)acos(light_dot));
							light_dot = p_gauss_dist_norm(sec_shininess, acosTable[(int)(light_dot*LUT_SIZE)]);

							sum_specular_shd.r += light_col.l_color_shd.r * light_dot * sec_specular.r * view_dot;
							sum_specular_shd.g += light_col.l_color_shd.g * light_dot * sec_specular.g * view_dot;
							sum_specular_shd.b += light_col.l_color_shd.b * light_dot * sec_specular.b * view_dot;

							// compute TRANSLUCENT specular highlight

							// recompute halfway vector, since we now base it on the
							// inverted view direction

							mi_vector_add(&h, &state->dir, &light_dir);
							mi_vector_normalize(&h);

							norm.x = h.x + base_tang.x * tran_deviation;
							norm.y = h.y + base_tang.y * tran_deviation;
							norm.z = h.z + base_tang.z * tran_deviation;

							mi_vector_normalize(&norm);
							light_dot = mi_vector_dot(&tang, &norm);
							light_dot = 1.0f - light_dot*light_dot;

		//					light_dot = p_gauss_dist_norm(tran_shininess, (miScalar)acos(light_dot)) * dim_fresnel;
		//					light_dot = p_gauss_dist_norm(1.0f / tran_shininess, acosTable[(int)(light_dot*LUT_SIZE)]) * dim_fresnel;
							light_dot = p_gauss_dist_norm(tran_shininess, acosTable[(int)(light_dot*LUT_SIZE)]) * dim_fresnel;

							// specular translucent
							sum_specular_shd.r += light_col.l_color_shd.r * light_dot * tran_specular.r * back_dot;
							sum_specular_shd.g += light_col.l_color_shd.g * light_dot * tran_specular.g * back_dot;
							sum_specular_shd.b += light_col.l_color_shd.b * light_dot * tran_specular.b * back_dot;
						}
					}
		//			lightTK_pt = NULL;
				}		
				if(lightTK_pt)
					mi_shaderstate_set(state, lightsTK_key, NULL, 0, miSS_LIFETIME_EYERAY);
				// add it in
				samples = iter->get_number_of_samples();
				if (samples) 
				{
					float inv_samples = 1.f/samples;
					if(diffuse_en)
					{
						sum_light_shd.r *= inv_samples;
						sum_light_shd.g *= inv_samples;
						sum_light_shd.b *= inv_samples;

						sum_light.r *= inv_samples;
						sum_light.g *= inv_samples;
						sum_light.b *= inv_samples;
					}
					if(specular_en)
					{
						sum_specular.r *= inv_samples;
						sum_specular.g *= inv_samples;
						sum_specular.b *= inv_samples;

						sum_specular_shd.r *= inv_samples;
						sum_specular_shd.g *= inv_samples;
						sum_specular_shd.b *= inv_samples;
					}
				}
				special.out_diffuse_shadow.r += sum_light_shd.r;
				special.out_diffuse_shadow.g += sum_light_shd.g;
				special.out_diffuse_shadow.b += sum_light_shd.b;
				special.out_diffuse.r += sum_light.r;
				special.out_diffuse.g += sum_light.g;
				special.out_diffuse.b += sum_light.b;
				special.out_specular_shadow.r += sum_specular_shd.r;
				special.out_specular_shadow.g += sum_specular_shd.g;
				special.out_specular_shadow.b += sum_specular_shd.b;
				special.out_specular.r += sum_specular.r;
				special.out_specular.g += sum_specular.g;
				special.out_specular.b += sum_specular.b;	

			}
		}

		if(diffuse_en)
		{
			special.out_diffuse_shadow.r *= diffuse_coef;
			special.out_diffuse_shadow.g *= diffuse_coef;
			special.out_diffuse_shadow.b *= diffuse_coef;
			special.out_diffuse.r *= diffuse_coef;
			special.out_diffuse.g *= diffuse_coef;
			special.out_diffuse.b *= diffuse_coef;
		}
		if(specular_en)
		{
			special.out_specular_shadow.r *= specular_coef;
			special.out_specular_shadow.g *= specular_coef;
			special.out_specular_shadow.b *= specular_coef;
			special.out_specular.r *= specular_coef;
			special.out_specular.g *= specular_coef;
			special.out_specular.b *= specular_coef;	
		}

//		if (0)
//		{
//			state->pri = priSave;
//			mi_ray_offset(state, &offset);
//		}

		//restore original normal
		state->normal = normal;


		if(sp->use_occlusion)									//Occlusion
			special.out_occlusion = *mi_eval_color(&paras->custom_occlusion);



		float coeffAmb = *mi_eval_scalar(&paras->cAmb);							//Ambient
		if(sp->use_ambient && (coeffAmb > miEPS))
		{
			miColor *amb_color = mi_eval_color(&paras->ambient_color);
			special.out_ambient.r = special.out_color.r * amb_color->r * coeffAmb;
			special.out_ambient.g = special.out_color.g * amb_color->g * coeffAmb;
			special.out_ambient.b = special.out_color.b * amb_color->b * coeffAmb;

												//Effect occlusion to ambient
			float effect_ambient = *mi_eval_scalar(&paras->effect_ambient);
			if(sp->use_occlusion && effect_ambient > miEPS)
			{
				if(effect_ambient == 1)
				{
					special.out_amb_occlusion.r = special.out_ambient.r * special.out_occlusion.r;
					special.out_amb_occlusion.g = special.out_ambient.g * special.out_occlusion.g;
					special.out_amb_occlusion.b = special.out_ambient.b * special.out_occlusion.b;
				}
				else
				{
					special.out_amb_occlusion.r = special.out_ambient.r * (special.out_occlusion.r * effect_ambient + (1 - effect_ambient));
					special.out_amb_occlusion.g = special.out_ambient.g * (special.out_occlusion.g * effect_ambient + (1 - effect_ambient));
					special.out_amb_occlusion.b = special.out_ambient.b * (special.out_occlusion.b * effect_ambient + (1 - effect_ambient));
				}
			}
			else
			{
				special.out_amb_occlusion.r = special.out_ambient.r;
				special.out_amb_occlusion.g = special.out_ambient.g;
				special.out_amb_occlusion.b = special.out_ambient.b;
			}
		}


		if(sp->use_indirect_illum)									//Indirect_Illum
		{
			miColor indirect_color;
			miColor * coeffInd;
			coeffInd = mi_eval_color(&paras->cInd);

			mi_compute_irradiance(&indirect_color, state);
			miScalar satur = *mi_eval_scalar(&paras->saturation_indirect);
	//		miScalar gray = color.r*0.299f + color.g*0.587f + color.b*0.114f;
			miScalar gray = (indirect_color.r + indirect_color.g + indirect_color.b) * 0.3333f;

			if(satur < 1)
			{
				if(satur == 0)
					indirect_color.r = indirect_color.g = indirect_color.b = gray;
				else
				{
					indirect_color.r = gray*(1-satur) + indirect_color.r*satur;
					indirect_color.g = gray*(1-satur) + indirect_color.g*satur;
					indirect_color.b = gray*(1-satur) + indirect_color.b*satur;
				}
			}

			
			float effect_indir = *mi_eval_scalar(&paras->effect_indirect);			//Effect occlusion to indirect
			if(sp->use_occlusion && effect_indir > miEPS)
			{
				if(effect_indir == 1)
				{
					special.out_indirect_illum.r = indirect_color.r * special.out_color.r * coeffInd->r * special.out_occlusion.r;
					special.out_indirect_illum.g = indirect_color.g * special.out_color.g * coeffInd->g * special.out_occlusion.g;
					special.out_indirect_illum.b = indirect_color.b * special.out_color.b * coeffInd->b * special.out_occlusion.b;
				}
				else
				{
					special.out_indirect_illum.r = special.out_color.r * coeffInd->r * (indirect_color.r * special.out_occlusion.r * effect_indir + (1 - effect_indir) * indirect_color.r);
					special.out_indirect_illum.g = special.out_color.g * coeffInd->g * (indirect_color.g * special.out_occlusion.g * effect_indir + (1 - effect_indir) * indirect_color.g);
					special.out_indirect_illum.b = special.out_color.b * coeffInd->b * (indirect_color.b * special.out_occlusion.b * effect_indir + (1 - effect_indir) * indirect_color.b);
				}
			}
			else
			{
				special.out_indirect_illum.r = indirect_color.r * special.out_color.r * coeffInd->r;
				special.out_indirect_illum.g = indirect_color.g * special.out_color.g * coeffInd->g;
				special.out_indirect_illum.b = indirect_color.b * special.out_color.b * coeffInd->b;
			}
		}


													//Effect occlusion to specular
		float effect_spec = *mi_eval_scalar(&paras->effect_specular);
		if(sp->use_occlusion && effect_spec > miEPS)
		{
			if(effect_spec == 1)
			{
				special.out_specular_shadow.r *= special.out_occlusion.r;
				special.out_specular_shadow.g *= special.out_occlusion.g;
				special.out_specular_shadow.b *= special.out_occlusion.b;
			}
			else
			{
				special.out_specular_shadow.r = special.out_specular_shadow.r * special.out_occlusion.r * effect_spec + (1 - effect_spec) * special.out_specular_shadow.r;
				special.out_specular_shadow.g = special.out_specular_shadow.g * special.out_occlusion.g * effect_spec + (1 - effect_spec) * special.out_specular_shadow.g;
				special.out_specular_shadow.b = special.out_specular_shadow.b * special.out_occlusion.b * effect_spec + (1 - effect_spec) * special.out_specular_shadow.b;
			}
		}


														//Effect occlusion to diffuse
		float effect_diffuse = *mi_eval_scalar(&paras->effect_diffuse);
		if(sp->use_occlusion && (effect_diffuse > miEPS))
		{
			if(effect_diffuse == 1)
			{
				special.out_diff_occlusion.r = special.out_color.r * special.out_occlusion.r;
				special.out_diff_occlusion.g = special.out_color.g * special.out_occlusion.g;
				special.out_diff_occlusion.b = special.out_color.b * special.out_occlusion.b;

//				special.out_diff_occlusion.r = special.out_occlusion.r;
//				special.out_diff_occlusion.g = special.out_occlusion.g;
//				special.out_diff_occlusion.b = special.out_occlusion.b;
			}
			else
			{
				special.out_occlusion.r = special.out_occlusion.r * effect_diffuse + (1 - effect_diffuse);
				special.out_occlusion.g = special.out_occlusion.g * effect_diffuse + (1 - effect_diffuse);
				special.out_occlusion.b = special.out_occlusion.b * effect_diffuse + (1 - effect_diffuse);

				special.out_diff_occlusion.r = special.out_color.r * special.out_occlusion.r;
				special.out_diff_occlusion.g = special.out_color.g * special.out_occlusion.g;
				special.out_diff_occlusion.b = special.out_color.b * special.out_occlusion.b;
			}
		}
		else
		{			
			special.out_diff_occlusion.r = special.out_color.r;
			special.out_diff_occlusion.g = special.out_color.g;
			special.out_diff_occlusion.b = special.out_color.b;
			special.out_occlusion.r = special.out_occlusion.g = special.out_occlusion.b = 1;
		}


		miScalar coeffInc = *mi_eval_scalar(&paras->cInc);										//Incandescence
		if(sp->use_incandescence && coeffInc != 0)
		{
			special.out_incandescence = *mi_eval_color(&paras->incandescence);
			special.out_incandescence.r *= coeffInc;
			special.out_incandescence.g *= coeffInc;
			special.out_incandescence.b *= coeffInc;
		}

	} //end of shading


	if(trans_en && state->options->scanline != 'r' && alpha <= miAlphaEPS_INV)
	{
		mi_shaderstate_set(state, "p_hair_stat", &mh_data, sizeof(mh_data), miSS_LIFETIME_EYERAY);
		miColor col = {0.0, 0.0, 0.0, 0.0};
		state->refraction_level--;
		mi_trace_transparent(&col, state);
		state->refraction_level++;
		result->r = inv_alpha * col.r;
		result->g = inv_alpha * col.g;
		result->b = inv_alpha * col.b;
		result->a += inv_alpha * col.a;
//		state->refraction_level++;
//		p_color_mix(result, result, &transparent, &trans_col);
	}

	if(mh_new && state->options->scanline != 'r')
		mi_shaderstate_set(state, "p_hair_stat", NULL, 0, miSS_LIFETIME_EYERAY);



//	if(final_gather_mode(state)) //it's final gather stage or shader called by final gather ray
//	{
//		irradiance = *mi_eval_color(&paras->m_irradiance);
//		p_color_mul(result, result, &irradiance);
//	}


	miScalar near = *mi_eval_scalar(&paras->near);
	miScalar far = *mi_eval_scalar(&paras->far);
	special.out_z.r = special.out_z.g = special.out_z.b = special.out_z.a = 0;
	if(*mi_eval_boolean(&paras->use_z) == miTRUE)																						//Z
	{
		miScalar depth = (miScalar)state->dist;
		miScalar aver = 1;

		if(*mi_eval_boolean(&paras->use_float_z) == miFALSE)
			aver = far - near;

		switch(state->type)
		{
			case miRAY_EYE:
				if(depth < near)
					depth = near;
				if(depth > far)
					depth = far;
				depth = (depth - near)/aver;
				special.out_z.r = depth;
				break;
			case miRAY_TRANSPARENT:
				if(*mi_eval_boolean(&paras->opacity_z) == miTRUE)
				{
					if(depth > far)
						depth = far;
					depth = depth / aver;
					special.out_z.r = depth;
				}
				break;
			default:
				break;
		}
	}

	if(result->a > 1)
		result->a = 1;
	else
	{
		if(result->a < 0)
			result->a = 0;
	}


	special.out_shading.r = special.out_diffuse_shadow.r * special.out_diff_occlusion.r;			//Shading
	special.out_shading.g = special.out_diffuse_shadow.g * special.out_diff_occlusion.g;
	special.out_shading.b = special.out_diffuse_shadow.b * special.out_diff_occlusion.b;

	//Result
	if(alpha > miAlphaEPS)
	{
		result->r += (special.out_shading.r + special.out_specular_shadow.r + special.out_amb_occlusion.r + special.out_incandescence.r + special.out_indirect_illum.r) * alpha;
		result->g += (special.out_shading.g + special.out_specular_shadow.g + special.out_amb_occlusion.g + special.out_incandescence.g + special.out_indirect_illum.g) * alpha;
		result->b += (special.out_shading.b + special.out_specular_shadow.b + special.out_amb_occlusion.b + special.out_incandescence.b + special.out_indirect_illum.b) * alpha;
//		result->a += alpha;

		if(sp->use_SpecialC1)											//SpecialC1
			special.out_specialC1 = *mi_eval_color(&paras->specialC1);
		if(sp->use_SpecialC2)											//SpecialC2
			special.out_specialC2 = *mi_eval_color(&paras->specialC2);
		if(sp->use_SpecialC3)											//SpecialC3
			special.out_specialC3 = *mi_eval_color(&paras->specialC3);
		if(sp->use_SpecialC4)											//SpecialC4
			special.out_specialC4 = *mi_eval_color(&paras->specialC4);
		if(sp->use_SpecialC5)											//SpecialC5
			special.out_specialC5 = *mi_eval_color(&paras->specialC5);
		if(sp->use_SpecialC6)											//SpecialC6
			special.out_specialC6 = *mi_eval_color(&paras->specialC6);
		if(sp->use_SpecialC7)											//SpecialC7
			special.out_specialC7 = *mi_eval_color(&paras->specialC7);
		if(sp->use_SpecialC8)											//SpecialC8
			special.out_specialC8 = *mi_eval_color(&paras->specialC8);
		if(sp->use_SpecialC9)											//SpecialC9
			special.out_specialC9 = *mi_eval_color(&paras->specialC9);
		if(sp->use_SpecialC10)											//SpecialC10
			special.out_specialC10 = *mi_eval_color(&paras->specialC10);
	}

		if(alpha <= miAlphaEPS_INV && state->options->scanline != 'r')
		{
			special.out_shading.r *= alpha;
			special.out_shading.g *= alpha;
			special.out_shading.b *= alpha;

			special.out_color.r *= alpha;
			special.out_color.g *= alpha;
			special.out_color.b *= alpha;

			special.out_diff_occlusion.r *= alpha;
			special.out_diff_occlusion.g *= alpha;
			special.out_diff_occlusion.b *= alpha;

			special.out_diffuse.r *= alpha;
			special.out_diffuse.g *= alpha;
			special.out_diffuse.b *= alpha;

			special.out_diffuse_shadow.r *= alpha;
			special.out_diffuse_shadow.g *= alpha;
			special.out_diffuse_shadow.b *= alpha;

			special.out_specular.r *= alpha;
			special.out_specular.g *= alpha;
			special.out_specular.b *= alpha;

			special.out_specular_shadow.r *= alpha;
			special.out_specular_shadow.g *= alpha;
			special.out_specular_shadow.b *= alpha;

			special.out_occlusion.r *= alpha;
			special.out_occlusion.g *= alpha;
			special.out_occlusion.b *= alpha;

			special.out_ambient.r *= alpha;
			special.out_ambient.g *= alpha;
			special.out_ambient.b *= alpha;

			special.out_amb_occlusion.r *= alpha;
			special.out_amb_occlusion.g *= alpha;
			special.out_amb_occlusion.b *= alpha;

			special.out_indirect_illum.r *= alpha;
			special.out_indirect_illum.g *= alpha;
			special.out_indirect_illum.b *= alpha;

			special.out_incandescence.r *= alpha;
			special.out_incandescence.g *= alpha;
			special.out_incandescence.b *= alpha;

			special.out_z.g *= alpha;
			special.out_z.b *= alpha;

			special.out_specialC1.r *= alpha;
			special.out_specialC1.g *= alpha;
			special.out_specialC1.b *= alpha;

			special.out_specialC2.r *= alpha;
			special.out_specialC2.g *= alpha;
			special.out_specialC2.b *= alpha;

			special.out_specialC3.r *= alpha;
			special.out_specialC3.g *= alpha;
			special.out_specialC3.b *= alpha;

			special.out_specialC4.r *= alpha;
			special.out_specialC4.g *= alpha;
			special.out_specialC4.b *= alpha;

			special.out_specialC5.r *= alpha;
			special.out_specialC5.g *= alpha;
			special.out_specialC5.b *= alpha;

			special.out_specialC6.r *= alpha;
			special.out_specialC6.g *= alpha;
			special.out_specialC6.b *= alpha;

			special.out_specialC7.r *= alpha;
			special.out_specialC7.g *= alpha;
			special.out_specialC7.b *= alpha;

			special.out_specialC8.r *= alpha;
			special.out_specialC8.g *= alpha;
			special.out_specialC8.b *= alpha;

			special.out_specialC9.r *= alpha;
			special.out_specialC9.g *= alpha;
			special.out_specialC9.b *= alpha;

			special.out_specialC10.r *= alpha;
			special.out_specialC10.g *= alpha;
			special.out_specialC10.b *= alpha;
		}

		if((state->type == miRAY_EYE) || (state->type == miRAY_TRANSPARENT))			//Writing buffers
		{
			mi::shader::Access_fb fbc(state->camera->buffertag);
			size_t index = -1;
			miColor temp_color;

			if(sp->shading_pass)
			{
				if(sp->shading_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_shading", index)) 
						sp->shading_buf = (int)index;
					else
						sp->shading_pass = miFALSE;				
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->shading_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_shading.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_shading.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_shading.b;
				mi_fb_put(state, sp->shading_buf, &temp_color);
			}
			if(sp->color_pass)
			{
				if(sp->color_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_color", index)) 
						sp->color_buf = (int)index;
					else
						sp->color_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->color_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_diff_occlusion.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_diff_occlusion.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_diff_occlusion.b;
				mi_fb_put(state, sp->color_buf, &temp_color);
			}
			if(sp->diffuse_pass)
			{
				if(sp->diffuse_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_diffuse", index)) 
						sp->diffuse_buf = (int)index;
					else
						sp->diffuse_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->diffuse_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_diffuse.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_diffuse.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_diffuse.b;
				mi_fb_put(state, sp->diffuse_buf, &temp_color);
			}
			if(sp->diffuse_shadow_pass)
			{
				if(sp->diffuse_shadow_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_shadow", index)) 
						sp->diffuse_shadow_buf  = (int)index;
					else
						sp->diffuse_shadow_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->diffuse_shadow_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_diffuse_shadow.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_diffuse_shadow.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_diffuse_shadow.b;
				mi_fb_put(state, sp->diffuse_shadow_buf, &temp_color);
			}
			if(sp->specular_pass)
			{
				if(sp->specular_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_specular", index)) 
						sp->specular_buf = (int)index;
					else
						sp->specular_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specular_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specular.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specular.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specular.b;
				mi_fb_put(state, sp->specular_buf, &temp_color);
			}
			if(sp->specular_shadow_pass)
			{
				if(sp->specular_shadow_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_spec_shd", index)) 
						sp->specular_shadow_buf = (int)index;
					else
						sp->specular_shadow_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specular_shadow_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specular_shadow.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specular_shadow.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specular_shadow.b;
				mi_fb_put(state, sp->specular_shadow_buf, &temp_color);
			}
			if(sp->diff_occlusion_pass)
			{
				if(sp->diff_occlusion_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_diff_occl", index))
						sp->diff_occlusion_buf = (int)index;
					else
						sp->diff_occlusion_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->diff_occlusion_buf, &temp_color);
				if(alpha > miEPS)
				{
					temp_color.r = temp_color.r * inv_alpha + special.out_occlusion.r;
					temp_color.g = temp_color.g * inv_alpha + special.out_occlusion.g;
					temp_color.b = temp_color.b * inv_alpha + special.out_occlusion.b;
				}
				mi_fb_put(state, sp->diff_occlusion_buf, &temp_color);
			}
			if(sp->ambient_pass)
			{
				if(sp->ambient_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_ambient", index)) 
						sp->ambient_buf = (int)index;
					else
						sp->ambient_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->ambient_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_ambient.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_ambient.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_ambient.b;
				mi_fb_put(state, sp->ambient_buf, &temp_color);
			}
			if(sp->amb_occlusion_pass)
			{
				if(sp->amb_occlusion_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_amb_occlus", index))
						sp->amb_occlusion_buf = (int)index;
					else
						sp->amb_occlusion_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->amb_occlusion_buf, &temp_color);
				if(alpha > miEPS)
				{
					temp_color.r = temp_color.r * inv_alpha + special.out_amb_occlusion.r;
					temp_color.g = temp_color.g * inv_alpha + special.out_amb_occlusion.g;
					temp_color.b = temp_color.b * inv_alpha + special.out_amb_occlusion.b;
				}
				mi_fb_put(state, sp->amb_occlusion_buf, &temp_color);
			}
			if(sp->indirect_illum_pass)
			{
				if(sp->indirect_illum_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_indirect", index)) 
						sp->indirect_illum_buf = (int)index;
					else
						sp->indirect_illum_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->indirect_illum_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_indirect_illum.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_indirect_illum.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_indirect_illum.b;
				mi_fb_put(state, sp->indirect_illum_buf, &temp_color);
			}


			if(sp->incandescence_pass)
			{
				if(sp->incandescence_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_incand", index)) 
						sp->incandescence_buf = (int)index;
					else
						sp->incandescence_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->incandescence_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_incandescence.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_incandescence.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_incandescence.b;
				mi_fb_put(state, sp->incandescence_buf, &temp_color);
			}
//			if(sp->bent_normal_pass)
//			{
//				index = -1;
//				fbc->get_index("tk_bent", index); 
//				sp->bent_normal_buf = (int)index;
//				miColor temp_bent_normal;
//				mi_fb_get(state, sp->bent_normal_buf, &temp_bent_normal);
//				temp_bent_normal.r = temp_bent_normal.r * inv_alpha + special.out_bent_normal.r;
//				temp_bent_normal.g = temp_bent_normal.g * inv_alpha + special.out_bent_normal.g;
//				temp_bent_normal.b = temp_bent_normal.b * inv_alpha + special.out_bent_normal.b;
//				mi_fb_put(state, sp->bent_normal_buf, &temp_bent_normal);
//			}
			if(sp->specialC1_pass)
			{
				if(sp->specialC1_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_specialC1", index)) 
						sp->specialC1_buf = (int)index;
					else
						sp->specialC1_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specialC1_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specialC1.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specialC1.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specialC1.b;
				mi_fb_put(state, sp->specialC1_buf, &temp_color);
			}
			if(sp->specialC2_pass)
			{
				if(sp->specialC2_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_specialC2", index)) 
						sp->specialC2_buf = (int)index;
					else
						sp->specialC2_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specialC2_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specialC2.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specialC2.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specialC2.b;
				mi_fb_put(state, sp->specialC2_buf, &temp_color);
			}
			if(sp->specialC3_pass)
			{
				if(sp->specialC3_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_specialC3", index))
						sp->specialC3_buf = (int)index;
					else
						sp->specialC3_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specialC3_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specialC3.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specialC3.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specialC3.b;
				mi_fb_put(state, sp->specialC3_buf, &temp_color);
			}
			if(sp->specialC4_pass)
			{
				if(sp->specialC4_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_specialC4", index)) 
						sp->specialC4_buf = (int)index;
					else
						sp->specialC4_pass = miFALSE;

				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specialC4_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specialC4.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specialC4.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specialC4.b;
				mi_fb_put(state, sp->specialC4_buf, &temp_color);
			}
			if(sp->specialC5_pass)
			{
				if(sp->specialC5_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_specialC5", index))
						sp->specialC5_buf = (int)index;
					else
						sp->specialC5_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specialC5_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specialC5.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specialC5.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specialC5.b;
				mi_fb_put(state, sp->specialC5_buf, &temp_color);
			}
			if(sp->specialC6_pass)
			{
				if(sp->specialC6_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_specialC6", index)) 
						sp->specialC6_buf = (int)index;
					else
						sp->specialC6_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specialC6_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specialC6.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specialC6.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specialC6.b;
				mi_fb_put(state, sp->specialC6_buf, &temp_color);
			}
			if(sp->specialC7_pass)
			{
				if(sp->specialC7_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_specialC7", index)) 
						sp->specialC7_buf = (int)index;
					else
						sp->specialC7_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specialC7_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specialC7.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specialC7.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specialC7.b;
				mi_fb_put(state, sp->specialC7_buf, &temp_color);
			}
			if(sp->specialC8_pass)
			{
				if(sp->specialC8_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_specialC8", index)) 
						sp->specialC8_buf = (int)index;
					else
						sp->specialC8_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specialC8_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specialC8.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specialC8.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specialC8.b;
				mi_fb_put(state, sp->specialC8_buf, &temp_color);
			}
			if(sp->specialC9_pass)
			{
				if(sp->specialC9_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_specialC9", index)) 
						sp->specialC9_buf = (int)index;
					else
						sp->specialC9_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specialC9_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specialC9.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specialC9.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specialC9.b;
				mi_fb_put(state, sp->specialC9_buf, &temp_color);
			}
			if(sp->specialC10_pass)
			{
				if(sp->specialC10_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_specialC10", index))
						sp->specialC10_buf = (int)index;
					else
						sp->specialC10_pass = miFALSE;
				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->specialC10_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_specialC10.r;
				temp_color.g = temp_color.g * inv_alpha + special.out_specialC10.g;
				temp_color.b = temp_color.b * inv_alpha + special.out_specialC10.b;
				mi_fb_put(state, sp->specialC10_buf, &temp_color);
			}
			if(sp->z_pass)
			{
				if(sp->z_buf < 0)
				{
					index = -1;
					if(fbc->get_index("tk_z", index)) 
						sp->z_buf = (int)index;
					else
						sp->z_pass = miFALSE;

				}
				temp_color.r = temp_color.g = temp_color.b = temp_color.a = 0;
				mi_fb_get(state, sp->z_buf, &temp_color);
				temp_color.r = temp_color.r * inv_alpha + special.out_z.r;
				temp_color.g = temp_color.g + special.out_z.g;
				temp_color.b = temp_color.b + special.out_z.b;

				if(sp->invert_z == miTRUE)
				{
					if(state->type == miRAY_EYE)
					{
						if(sp->use_float_z == miFALSE)
						{
							temp_color.r = 1 -temp_color.r;
							temp_color.g = 1 -temp_color.g;
							temp_color.b = 1 -temp_color.b;
						}
						else
						{
							temp_color.r = (far - near) -temp_color.r;
							temp_color.g = (far - near) -temp_color.g;
							temp_color.b = (far - near) -temp_color.b;
						}
					}
				}
				temp_color.a += special.out_z.a;
				mi_fb_put(state, sp->z_buf, &temp_color);
			}	
		}
	
	//Set opacity for rasterizer
//	if(state->options->scanline == 'r')
//	{
//		miColor rasterizer_opacity;
//		rasterizer_opacity.r = rasterizer_opacity.g = rasterizer_opacity.b = rasterizer_opacity.a = alpha;
//		mi_opacity_set(state, &rasterizer_opacity);
//	}

	//check for NaN
	if(result->r != result->r)
	{
		mi_warning("p_HairTK shader return NaN or infinite R value. Replaced with zero");
		result->r = 0;
	}
	if(result->g != result->g)
	{
		mi_warning("p_HairTK shader return NaN or infinite G value. Replaced with zero");
		result->g = 0;
	}
	if(result->b != result->b)
	{
		mi_warning("p_HairTK shader return NaN or infinite B value. Replaced with zero");
		result->b = 0;
	}
	if(result->a != result->a)
	{
		mi_warning("p_HairTK shader return NaN or infinite A value. Replaced with zero");
		result->a = 0;
	}
	return miTRUE;
}

//	Shader cleanup function
DLLEXPORT miBoolean p_HairTK_exit(miState *state, p_HairTK_t *paras)
{
	struct static_hair_param **spp;

	if (!paras)
	{
		mi_mem_release(cosTable);
		mi_mem_release(acosTable);
	}
	else {
		mi_query(miQ_FUNC_USERPTR, state, 0, &spp);
		mi_mem_release(*spp);
		*spp = NULL;
	}
	return(miTRUE);
}
}
