//  Copyright (c) 2006, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <math.h>
#include <shader.h>
#include <ImfOutputFile.h>
#include <ImfHeader.h>
#include <ImfChannelList.h>
#include <ImfArray.h>
#include <ImfConvert.h>
#include <ImfPixelType.h>
#include <ImfChannelListAttribute.h>
#include <ImfStandardAttributes.h>
#include <ImfMatrixAttribute.h>
#include "geoshader.h"
#include "p_utils.h"

typedef struct
{
	miTag		fileName;
	int			compression;
	miBoolean	rgbaPass;
	int			rgbaFormat;
	miBoolean	zPass;
	int			zFormat;
	miBoolean	nPass;
	int			nFormat;
	miBoolean	mPass;
	int			mFormat;
	miBoolean	tPass;
	int			tFormat;
	miBoolean	cPass;
	int			cFormat;
	miBoolean 	shadingPass;
	miTag			shadingBuffer;
	int			shadingFormat;
	miBoolean 	colorPass;
	miTag			colorBuffer;
	int			colorFormat;
	miBoolean 	diffusePass;
	miTag			diffuseBuffer;
	int			diffuseFormat;
	miBoolean 	diffuseShadowPass;
	miTag			diffuseShadowBuffer;
	int			diffuseShadowFormat;
	miBoolean 	specularPass;
	miTag			specularBuffer;
	int			specularFormat;
	miBoolean 	specularShadowPass;
	miTag			specularShadowBuffer;
	int			specularShadowFormat;
	miBoolean 	diffuseOcclusionPass;
	miTag			diffuseOcclusionBuffer;
	int			diffuseOcclusionFormat;
	miBoolean 	indirectIllumPass;
	miTag			indirectIllumBuffer;
	int			indirectIllumFormat;
	miBoolean 	ambientPass;
	miTag			ambientBuffer;
	int			ambientFormat;
	miBoolean 	ambOcclusionPass;
	miTag			ambOcclusionBuffer;
	int			ambOcclusionFormat;
	miBoolean 	reflectPass;
	miTag			reflectBuffer;
	int			reflectFormat;
	miBoolean 	refractPass;
	miTag			refractBuffer;
	int			refractFormat;
	miBoolean 	incandescencePass;
	miTag			incandescenceBuffer;
	int			incandescenceFormat;
	miBoolean 	depthPass;
	miTag			depthBuffer;
	int			depthFormat;
	miBoolean 	bentNormalPass;
	miTag			bentNormalBuffer;
	int			bentNormalFormat;
	miBoolean 	specialC1Pass;
	miTag			specialC1Buffer;
	int			specialC1Format;
	miBoolean 	specialC2Pass;
	miTag			specialC2Buffer;
	int			specialC2Format;
	miBoolean 	specialC3Pass;
	miTag			specialC3Buffer;
	int			specialC3Format;
	miBoolean 	specialC4Pass;
	miTag			specialC4Buffer;
	int			specialC4Format;
	miBoolean 	specialC5Pass;
	miTag			specialC5Buffer;
	int			specialC5Format;
	miBoolean 	specialC6Pass;
	miTag			specialC6Buffer;
	int			specialC6Format;
	miBoolean 	specialC7Pass;
	miTag			specialC7Buffer;
	int			specialC7Format;
	miBoolean 	specialC8Pass;
	miTag			specialC8Buffer;
	int			specialC8Format;
	miBoolean 	specialC9Pass;
	miTag			specialC9Buffer;
	int			specialC9Format;
	miBoolean 	specialC10Pass;
	miTag			specialC10Buffer;
	int			specialC10Format;
}p_OpenEXR_saver_t;

typedef struct
{
	float		*floatR;
	float		*floatG;
	float		*floatB;
	half		*halfR;
	half		*halfG;
	half		*halfB;
	unsigned int	*unintR;
	unsigned int	*unintG;
	unsigned int	*unintB;
}pass_pt;

int save_pass(miState *state, int buf_num, int format, int width, int height, char *name, Imf::Header *exrHeader, Imf::FrameBuffer *exrFrameBuffer, pass_pt * data)
{
	miColor color;
	char channel_name_R[32];
	char channel_name_G[32];
	char channel_name_B[32];
	strcpy(channel_name_R, name);
	strcpy(channel_name_G, name);
	strcpy(channel_name_B, name);
	strcat(channel_name_R, ".r");
	strcat(channel_name_G, ".g");
	strcat(channel_name_B, ".b");
	switch (format)
	{
		case 0:
		{
			exrHeader->channels().insert (channel_name_R, Imf::Channel(Imf::UINT));
			exrHeader->channels().insert (channel_name_G, Imf::Channel(Imf::UINT));
			exrHeader->channels().insert (channel_name_B, Imf::Channel(Imf::UINT));

			miImg_image *fb_pass = mi_output_image_open(state, buf_num);

//			data->unintR = new unsigned int[width*height];
//			data->unintG = new unsigned int[width*height];
//			data->unintB = new unsigned int[width*height];

			data->unintR = (unsigned int *)mi_mem_test_allocate((sizeof(unsigned int))*width*height);
			if(data->unintR == 0)
			{
				mi_output_image_close(state, buf_num);
				return 0;
			}
			data->unintG = (unsigned int *)mi_mem_test_allocate((sizeof(unsigned int))*width*height);
			if(data->unintG == 0)
			{
				mi_mem_release(data->unintR);
				mi_output_image_close(state, buf_num);
				return 0;
			}
			data->unintB = (unsigned int *)mi_mem_test_allocate((sizeof(unsigned int))*width*height);
			if(data->unintB == 0)
			{
				mi_mem_release(data->unintR);
				mi_mem_release(data->unintG);
				mi_output_image_close(state, buf_num);
				return 0;
			}

			for (int x = 0; x < width; x++)
			{
				if (mi_par_aborted())
					break;
				for (int y = 0; y < height; y++)
				{
					mi_img_get_color(fb_pass, &color, x, height-y-1);
					data->unintR[x+y*width] = Imf::floatToUint(color.r);
					data->unintG[x+y*width] = Imf::floatToUint(color.g);
					data->unintB[x+y*width] = Imf::floatToUint(color.b);
				}
			}
			mi_output_image_close(state, buf_num);
			mi_progress("p_OpenEXR_saver(): %s...(unint)", name);
			exrFrameBuffer->insert(channel_name_R, Imf::Slice(Imf::UINT, (char*)(&data->unintR[0]), sizeof(unsigned int)*1, sizeof(unsigned int)*width));
			exrFrameBuffer->insert(channel_name_G, Imf::Slice(Imf::UINT, (char*)(&data->unintG[0]), sizeof(unsigned int)*1, sizeof(unsigned int)*width));
			exrFrameBuffer->insert(channel_name_B, Imf::Slice(Imf::UINT, (char*)(&data->unintB[0]), sizeof(unsigned int)*1, sizeof(unsigned int)*width));
			break;
		}
		case 1:
		{
			exrHeader->channels().insert (channel_name_R, Imf::Channel(Imf::HALF));
			exrHeader->channels().insert (channel_name_G, Imf::Channel(Imf::HALF));
			exrHeader->channels().insert (channel_name_B, Imf::Channel(Imf::HALF));

			miImg_image *fb_pass = mi_output_image_open(state, buf_num);

//			data->halfR = new half[width*height];
//			data->halfG = new half[width*height];
//			data->halfB = new half[width*height];

			data->halfR = (half *)mi_mem_test_allocate((sizeof(half))*width*height);
			if(data->halfR == 0)
			{
				mi_output_image_close(state, buf_num);
				return 0;
			}
			data->halfG = (half *)mi_mem_test_allocate((sizeof(half))*width*height);
			if(data->halfG == 0)
			{
				mi_mem_release(data->halfR);
				mi_output_image_close(state, buf_num);
				return 0;
			}
			data->halfB = (half *)mi_mem_test_allocate((sizeof(half))*width*height);
			if(data->halfB == 0)
			{
				mi_mem_release(data->halfR);
				mi_mem_release(data->halfG);
				mi_output_image_close(state, buf_num);
				return 0;
			}

			for (int x = 0; x < width; x++)
			{
				if (mi_par_aborted())
					break;
				for (int y = 0; y < height; y++)
				{
					mi_img_get_color(fb_pass, &color, x, height-y-1);
					data->halfR[x+y*width] = Imf::floatToHalf(color.r);
					data->halfG[x+y*width] = Imf::floatToHalf(color.g);
					data->halfB[x+y*width] = Imf::floatToHalf(color.b);
				}
			}
			mi_output_image_close(state, buf_num);
			mi_progress("p_OpenEXR_saver(): %s...(half)", name);
			exrFrameBuffer->insert(channel_name_R, Imf::Slice(Imf::HALF, (char*)(&data->halfR[0]), sizeof(half)*1, sizeof(half)*width));
			exrFrameBuffer->insert(channel_name_G, Imf::Slice(Imf::HALF, (char*)(&data->halfG[0]), sizeof(half)*1, sizeof(half)*width));
			exrFrameBuffer->insert(channel_name_B, Imf::Slice(Imf::HALF, (char*)(&data->halfB[0]), sizeof(half)*1, sizeof(half)*width));
			break;
		}
		case 2:
		{
			exrHeader->channels().insert (channel_name_R, Imf::Channel(Imf::FLOAT));
			exrHeader->channels().insert (channel_name_G, Imf::Channel(Imf::FLOAT));
			exrHeader->channels().insert (channel_name_B, Imf::Channel(Imf::FLOAT));

			miImg_image *fb_pass = mi_output_image_open(state, buf_num);

//			data->floatR = new float[width*height];
//			data->floatG = new float[width*height];
//			data->floatB = new float[width*height];

			data->floatR = (float *)mi_mem_test_allocate((sizeof(float))*width*height);
			if(data->floatR == 0)
			{
				mi_output_image_close(state, buf_num);
				return 0;
			}
			data->floatG = (float *)mi_mem_test_allocate((sizeof(float))*width*height);
			if(data->floatG == 0)
			{
				mi_mem_release(data->floatR);
				mi_output_image_close(state, buf_num);
				return 0;
			}
			data->floatB = (float *)mi_mem_test_allocate((sizeof(float))*width*height);
			if(data->floatB == 0)
			{
				mi_mem_release(data->floatR);
				mi_mem_release(data->floatG);
				mi_output_image_close(state, buf_num);
				return 0;
			}

			for (int x = 0; x < width; x++)
			{
				if (mi_par_aborted())
					break;
				for (int y = 0; y < height; y++)
				{
					mi_img_get_color(fb_pass, &color, x, height-y-1);
					data->floatR[x+y*width] = color.r;
					data->floatG[x+y*width] = color.g;
					data->floatB[x+y*width] = color.b;
				}
			}
			mi_output_image_close(state, buf_num);
			mi_progress("p_OpenEXR_saver(): %s...(float)", name);
			exrFrameBuffer->insert(channel_name_R, Imf::Slice(Imf::FLOAT, (char*)(&data->floatR[0]), sizeof(float)*1, sizeof(float)*width));
			exrFrameBuffer->insert(channel_name_G, Imf::Slice(Imf::FLOAT, (char*)(&data->floatG[0]), sizeof(float)*1, sizeof(float)*width));
			exrFrameBuffer->insert(channel_name_B, Imf::Slice(Imf::FLOAT, (char*)(&data->floatB[0]), sizeof(float)*1, sizeof(float)*width));
			break;
		}
	}
	return 1;
}

int get_fb_num_by_name(miState *state, miTag buf_tag)
{
	int result = -1;
	if (buf_tag == miNULLTAG)
	{
		mi_error("OpenEXR_saver shader: could not eval buffer miTag");
		return result;
	}
	char *buffer_name = (char*)mi_db_access(buf_tag);
	size_t index = -1;
	mi::shader::Access_fb fbc(state->camera->buffertag);
	fbc->get_index(buffer_name, index);
	result = (int)index;
	mi_db_unpin(buf_tag);
	return result;
}

extern "C"
{
DLLEXPORT int p_OpenEXR_saver_version(void){ return 3;}

DLLEXPORT miBoolean p_OpenEXR_saver(void *result, miState *state, p_OpenEXR_saver_t *paras)
{
//	Imf::PixelType	bytes;

	miTag		fileNameTag;
	char		*fileName;
	int		compression;

	miBoolean	rgbaPass = *mi_eval_boolean(&paras->rgbaPass);
	int			rgbaFormat = *mi_eval_integer(&paras->rgbaFormat);
	miBoolean	zPass = *mi_eval_boolean(&paras->zPass);
	int			zFormat = *mi_eval_integer(&paras->zFormat);
	miBoolean	nPass = *mi_eval_boolean(&paras->nPass);
	int			nFormat = *mi_eval_integer(&paras->nFormat);
	miBoolean	mPass = *mi_eval_boolean(&paras->mPass);
	int			mFormat = *mi_eval_integer(&paras->mFormat);
	miBoolean	tPass = *mi_eval_boolean(&paras->tPass);
	int			tFormat = *mi_eval_integer(&paras->tFormat);
	miBoolean	cPass = *mi_eval_boolean(&paras->cPass);
	int			cFormat = *mi_eval_integer(&paras->cFormat);

	miBoolean	shadingPass = *mi_eval_boolean(&paras->shadingPass);
	int			shadingFormat = *mi_eval_integer(&paras->shadingFormat);
	int			shadingBuffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->shadingBuffer));

	miBoolean	colorPass = *mi_eval_boolean(&paras->colorPass);
	int			colorFormat = *mi_eval_integer(&paras->colorFormat);
	int			colorBuffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->colorBuffer));

	miBoolean	diffusePass = *mi_eval_boolean(&paras->diffusePass);
	int			diffuseFormat = *mi_eval_integer(&paras->diffuseFormat);
	int			diffuseBuffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->diffuseBuffer));

	miBoolean	diffuseShadowPass = *mi_eval_boolean(&paras->diffuseShadowPass);
	int			diffuseShadowFormat = *mi_eval_integer(&paras->diffuseShadowFormat);
	int			diffuseShadowBuffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->diffuseShadowBuffer));

	miBoolean	specularPass = *mi_eval_boolean(&paras->specularPass);
	int			specularFormat = *mi_eval_integer(&paras->specularFormat);
	int			specularBuffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->specularBuffer));

	miBoolean	specularShadowPass = *mi_eval_boolean(&paras->specularShadowPass);
	int			specularShadowFormat = *mi_eval_integer(&paras->specularShadowFormat);
	int			specularShadowBuffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->specularShadowBuffer));

	miBoolean	diffuseOcclusionPass = *mi_eval_boolean(&paras->diffuseOcclusionPass);
	int			diffuseOcclusionFormat = *mi_eval_integer(&paras->diffuseOcclusionFormat);
	int			diffuseOcclusionBuffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->diffuseOcclusionBuffer));

	miBoolean	indirectIllumPass = *mi_eval_boolean(&paras->indirectIllumPass);
	int			indirectIllumFormat = *mi_eval_integer(&paras->indirectIllumFormat);
	int			indirectIllumBuffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->indirectIllumBuffer));

	miBoolean	ambientPass = *mi_eval_boolean(&paras->ambientPass);
	int			ambientFormat = *mi_eval_integer(&paras->ambientFormat);
	int			ambientBuffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->ambientBuffer));

	miBoolean	ambOcclusionPass = *mi_eval_boolean(&paras->ambOcclusionPass);
	int			ambOcclusionFormat = *mi_eval_integer(&paras->ambOcclusionFormat);
	int			ambOcclusionBuffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->ambOcclusionBuffer));

	miBoolean	reflectPass = *mi_eval_boolean(&paras->reflectPass);
	int			reflectFormat = *mi_eval_integer(&paras->reflectFormat);
	int			reflectBuffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->reflectBuffer));

	miBoolean	refractPass = *mi_eval_boolean(&paras->refractPass);
	int			refractFormat = *mi_eval_integer(&paras->refractFormat);
	int			refractBuffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->refractBuffer));

	miBoolean	incandescencePass = *mi_eval_boolean(&paras->incandescencePass);
	int			incandescenceFormat = *mi_eval_integer(&paras->incandescenceFormat);
	int			incandescenceBuffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->incandescenceBuffer));

	miBoolean	depthPass = *mi_eval_boolean(&paras->depthPass);
	int			depthFormat = *mi_eval_integer(&paras->depthFormat);
	int			depthBuffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->depthBuffer));

	miBoolean	bentNormalPass = *mi_eval_boolean(&paras->bentNormalPass);
	int			bentNormalFormat = *mi_eval_integer(&paras->bentNormalFormat);
	int			bentNormalBuffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->bentNormalBuffer));

	miBoolean	specialC1Pass = *mi_eval_boolean(&paras->specialC1Pass);
	int			specialC1Format = *mi_eval_integer(&paras->specialC1Format);
	int			specialC1Buffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->specialC1Buffer));

	miBoolean	specialC2Pass = *mi_eval_boolean(&paras->specialC2Pass);
	int			specialC2Format = *mi_eval_integer(&paras->specialC2Format);
	int			specialC2Buffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->specialC2Buffer));

	miBoolean	specialC3Pass = *mi_eval_boolean(&paras->specialC3Pass);
	int			specialC3Format = *mi_eval_integer(&paras->specialC3Format);
	int			specialC3Buffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->specialC3Buffer));

	miBoolean	specialC4Pass = *mi_eval_boolean(&paras->specialC4Pass);
	int			specialC4Format = *mi_eval_integer(&paras->specialC4Format);
	int			specialC4Buffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->specialC4Buffer));		

	miBoolean	specialC5Pass = *mi_eval_boolean(&paras->specialC5Pass);
	int			specialC5Format = *mi_eval_integer(&paras->specialC5Format);
	int			specialC5Buffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->specialC5Buffer));

	miBoolean	specialC6Pass = *mi_eval_boolean(&paras->specialC6Pass);
	int			specialC6Format = *mi_eval_integer(&paras->specialC6Format);
	int			specialC6Buffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->specialC6Buffer));

	miBoolean	specialC7Pass = *mi_eval_boolean(&paras->specialC7Pass);
	int			specialC7Format = *mi_eval_integer(&paras->specialC7Format);
	int			specialC7Buffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->specialC7Buffer));

	miBoolean	specialC8Pass = *mi_eval_boolean(&paras->specialC8Pass);
	int			specialC8Format = *mi_eval_integer(&paras->specialC8Format);
	int			specialC8Buffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->specialC8Buffer));

	miBoolean	specialC9Pass = *mi_eval_boolean(&paras->specialC9Pass);
	int			specialC9Format = *mi_eval_integer(&paras->specialC9Format);
	int			specialC9Buffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->specialC9Buffer));

	miBoolean	specialC10Pass = *mi_eval_boolean(&paras->specialC10Pass);
	int			specialC10Format = *mi_eval_integer(&paras->specialC10Format);
	int			specialC10Buffer = miRC_IMAGE_USER + get_fb_num_by_name(state, *mi_eval_tag(&paras->specialC10Buffer));	

	miImg_image	*fb_rgbaPass;
	miImg_image	*fb_zPass;
	miImg_image	*fb_nPass;
	miImg_image	*fb_mPass;
	miImg_image	*fb_tPass;
	miImg_image	*fb_cPass;

	float		*rgbaRPixels32;
	float		*rgbaGPixels32;
	float		*rgbaBPixels32;
	float		*rgbaAPixels32;
	half		*rgbaRPixels16;
	half		*rgbaGPixels16;
	half		*rgbaBPixels16;
	half		*rgbaAPixels16;
	unsigned int	*rgbaRPixels8;
	unsigned int	*rgbaGPixels8;
	unsigned int	*rgbaBPixels8;
	unsigned int	*rgbaAPixels8;

	float		*depthZPixels32;
	half		*depthZPixels16;
	unsigned int	*depthZPixels8;

	float		*nRPixels32;
	float		*nGPixels32;
	float		*nBPixels32;
	half		*nRPixels16;
	half		*nGPixels16;
	half		*nBPixels16;
	unsigned int	*nRPixels8;
	unsigned int	*nGPixels8;
	unsigned int	*nBPixels8;

	float		*mRPixels32;
	float		*mGPixels32;
	float		*mBPixels32;
	half		*mRPixels16;
	half		*mGPixels16;
	half		*mBPixels16;
	unsigned int	*mRPixels8;
	unsigned int	*mGPixels8;
	unsigned int	*mBPixels8;

	float		*tZPixels32;
	half		*tZPixels16;
	unsigned int	*tZPixels8;

	float		*cZPixels32;
	half		*cZPixels16;
	unsigned int	*cZPixels8;

	miColor		color;
	miVector	mvector;
	miScalar	scalar;
	miUint		unsint;

	fileNameTag = *mi_eval_tag(&paras->fileName);
	if (!fileNameTag)
		return miFALSE;
	fileName = (char*)mi_db_access(fileNameTag);
	mi_db_unpin(fileNameTag);

	int		x, y;
	int		width, height;
	width = state->camera->x_resolution;
	height = state->camera->y_resolution;


	Imf::Header exrHeader(width, height);

	compression = *mi_eval_integer(&paras->compression)-1;
	switch(compression)
	{
		case 0:
			exrHeader.compression() = Imf::NO_COMPRESSION;
			break;
		case 1:
			exrHeader.compression() = Imf::RLE_COMPRESSION;
			break;
		case 2:
			exrHeader.compression() = Imf::ZIPS_COMPRESSION;
			break;
		case 3:
			exrHeader.compression() = Imf::ZIP_COMPRESSION;
			break;
		case 4:
			exrHeader.compression() = Imf::PIZ_COMPRESSION;
			break;
		case 5:
			exrHeader.compression() = Imf::PXR24_COMPRESSION;
			break;
		case 6:
			exrHeader.compression() = Imf::B44_COMPRESSION;
			break;
		case 7:
			exrHeader.compression() = Imf::B44A_COMPRESSION;
			break;
		default:
			exrHeader.compression() = Imf::NO_COMPRESSION;
			break;
	}

	mi_progress("p_OpenEXR_saver(): writing image file %s", fileName);

	Imf::FrameBuffer exrFrameBuffer;

//Processing RGBA channel
	if(rgbaPass == miTRUE)
	{
		switch (rgbaFormat)
		{
			case 0:			//UINT
			{
				exrHeader.channels().insert ("R", Imf::Channel(Imf::UINT));
				exrHeader.channels().insert ("G", Imf::Channel(Imf::UINT));
				exrHeader.channels().insert ("B", Imf::Channel(Imf::UINT));
				exrHeader.channels().insert ("A", Imf::Channel(Imf::UINT));

				fb_rgbaPass = mi_output_image_open(state, miRC_IMAGE_RGBA);

				rgbaRPixels8 = new unsigned int[width*height];
				rgbaGPixels8 = new unsigned int[width*height];
				rgbaBPixels8 = new unsigned int[width*height];
				rgbaAPixels8 = new unsigned int[width*height];
				for (x=0; x<width; x++)
				{
					if (mi_par_aborted())
						break;
					for (y=0; y<height; y++)
					{
						mi_img_get_color(fb_rgbaPass, &color, x, height-y-1);
						rgbaRPixels8[x+y*width] = Imf::floatToUint(color.r);
						rgbaGPixels8[x+y*width] = Imf::floatToUint(color.g);
						rgbaBPixels8[x+y*width] = Imf::floatToUint(color.b);
						rgbaAPixels8[x+y*width] = Imf::floatToUint(color.a);
					}
				}
			    mi_output_image_close(state, miRC_IMAGE_RGBA);
				mi_progress("p_OpenEXR_saver(): RGBA...(unint)");
				exrFrameBuffer.insert("R", Imf::Slice(Imf::UINT, (char*)(&rgbaRPixels8[0]), sizeof(unsigned int)*1, sizeof(unsigned int)*width));
				exrFrameBuffer.insert("G", Imf::Slice(Imf::UINT, (char*)(&rgbaGPixels8[0]), sizeof(unsigned int)*1, sizeof(unsigned int)*width));
				exrFrameBuffer.insert("B", Imf::Slice(Imf::UINT, (char*)(&rgbaBPixels8[0]), sizeof(unsigned int)*1, sizeof(unsigned int)*width));
				exrFrameBuffer.insert("A", Imf::Slice(Imf::UINT, (char*)(&rgbaAPixels8[0]), sizeof(unsigned int)*1, sizeof(unsigned int)*width));
				break;
			}
			case 1:				//HALF
			{
				exrHeader.channels().insert ("R", Imf::Channel(Imf::HALF));
				exrHeader.channels().insert ("G", Imf::Channel(Imf::HALF));
				exrHeader.channels().insert ("B", Imf::Channel(Imf::HALF));
				exrHeader.channels().insert ("A", Imf::Channel(Imf::HALF));

				fb_rgbaPass = mi_output_image_open(state, miRC_IMAGE_RGBA);

				rgbaRPixels16 = new half[width*height];
				rgbaGPixels16 = new half[width*height];
				rgbaBPixels16 = new half[width*height];
				rgbaAPixels16 = new half[width*height];

				for (x=0; x<width; x++)
				{
					if (mi_par_aborted())
						break;
					for (y=0; y<height; y++)
					{
						mi_img_get_color(fb_rgbaPass, &color, x, height-y-1);
						rgbaRPixels16[x+y*width] = Imf::floatToHalf(color.r);
						rgbaGPixels16[x+y*width] = Imf::floatToHalf(color.g);
						rgbaBPixels16[x+y*width] = Imf::floatToHalf(color.b);
						rgbaAPixels16[x+y*width] = Imf::floatToHalf(color.a);
					}
				}
			    mi_output_image_close(state, miRC_IMAGE_RGBA);
				mi_progress("p_OpenEXR_saver(): RGBA...(half)");
				exrFrameBuffer.insert("R", Imf::Slice(Imf::HALF, (char*)(&rgbaRPixels16[0]), sizeof(half)*1, sizeof(half)*width));
				exrFrameBuffer.insert("G", Imf::Slice(Imf::HALF, (char*)(&rgbaGPixels16[0]), sizeof(half)*1, sizeof(half)*width));
				exrFrameBuffer.insert("B", Imf::Slice(Imf::HALF, (char*)(&rgbaBPixels16[0]), sizeof(half)*1, sizeof(half)*width));
				exrFrameBuffer.insert("A", Imf::Slice(Imf::HALF, (char*)(&rgbaAPixels16[0]), sizeof(half)*1, sizeof(half)*width));
				break;
			}
			case 2:				//FLOAT
			{
				exrHeader.channels().insert ("R", Imf::Channel(Imf::FLOAT));
				exrHeader.channels().insert ("G", Imf::Channel(Imf::FLOAT));
				exrHeader.channels().insert ("B", Imf::Channel(Imf::FLOAT));
				exrHeader.channels().insert ("A", Imf::Channel(Imf::FLOAT));

				fb_rgbaPass = mi_output_image_open(state, miRC_IMAGE_RGBA);

				rgbaRPixels32 = new float[width*height];
				rgbaGPixels32 = new float[width*height];
				rgbaBPixels32 = new float[width*height];
				rgbaAPixels32 = new float[width*height];
				for (x = 0; x < width; x++)
				{
					if (mi_par_aborted())
						break;
					for (y = 0; y < height; y++)
					{
						mi_img_get_color(fb_rgbaPass, &color, x, height-y-1);
						rgbaRPixels32[x+y*width] = color.r;
						rgbaGPixels32[x+y*width] = color.g;
						rgbaBPixels32[x+y*width] = color.b;
						rgbaAPixels32[x+y*width] = color.a;
					}
				}
			    mi_output_image_close(state, miRC_IMAGE_RGBA);
				mi_progress("p_OpenEXR_saver(): RGBA...(float)");
				exrFrameBuffer.insert("R", Imf::Slice(Imf::FLOAT, (char*)(&rgbaRPixels32[0]), sizeof(float)*1, sizeof(float)*width));
				exrFrameBuffer.insert("G", Imf::Slice(Imf::FLOAT, (char*)(&rgbaGPixels32[0]), sizeof(float)*1, sizeof(float)*width));
				exrFrameBuffer.insert("B", Imf::Slice(Imf::FLOAT, (char*)(&rgbaBPixels32[0]), sizeof(float)*1, sizeof(float)*width));
				exrFrameBuffer.insert("A", Imf::Slice(Imf::FLOAT, (char*)(&rgbaAPixels32[0]), sizeof(float)*1, sizeof(float)*width));
				break;
			}
		}
	}


//Processing Z channel
	if(zPass == miTRUE)
	{
		switch (zFormat)
		{
			case 0:				//UINT
			{
				exrHeader.channels().insert ("Z.z", Imf::Channel(Imf::UINT));
			    fb_zPass = mi_output_image_open(state, miRC_IMAGE_Z);
				depthZPixels8 = new unsigned int[width*height];
				for (x=0; x<width; x++)
				{
					if (mi_par_aborted())
						break;
					for (y=0; y<height; y++)
					{
						mi_img_get_depth(fb_zPass, &scalar, x, height-y-1);
						depthZPixels8[x+y*width] = Imf::floatToUint(scalar);
					}
				}
			    mi_output_image_close(state, miRC_IMAGE_Z);
				mi_progress("p_OpenEXR_saver(): Z...(unint)");
				exrFrameBuffer.insert("Z.z", Imf::Slice(Imf::UINT, (char*)(&depthZPixels8[0]), sizeof(unsigned int)*1,	sizeof(unsigned int)*width));
				break;
			}
			case 1:				//HALF
			{
				exrHeader.channels().insert ("Z.z", Imf::Channel(Imf::HALF));
			    fb_zPass = mi_output_image_open(state, miRC_IMAGE_Z);
				depthZPixels16 = new half[width*height];
				for (x=0; x<width; x++)
				{
					if (mi_par_aborted())
						break;
					for (y=0; y<height; y++)
					{
						mi_img_get_depth(fb_zPass, &scalar, x, height-y-1);
						depthZPixels16[x+y*width] = Imf::floatToHalf(scalar);
					}
				}
				mi_output_image_close(state, miRC_IMAGE_Z);
				mi_progress("p_OpenEXR_saver(): Z...(half)");
				exrFrameBuffer.insert("Z.z", Imf::Slice(Imf::HALF, (char*)(&depthZPixels16[0]), sizeof(half)*1, sizeof(half)*width));
				break;
			}
			case 2:			//FLOAT
			{
				exrHeader.channels().insert ("Z.z", Imf::Channel(Imf::FLOAT));
			    fb_zPass = mi_output_image_open(state, miRC_IMAGE_Z);
				depthZPixels32 = new float[width*height];
				for (x = 0; x < width; x++)
				{
					if (mi_par_aborted())
						break;
					for (y = 0; y < height; y++)
					{
						mi_img_get_depth(fb_zPass, &scalar, x, height-y-1);
						depthZPixels32[x+y*width] = scalar;
					}
				}
				mi_output_image_close(state, miRC_IMAGE_Z);
				mi_progress("p_OpenEXR_saver(): Z...(float)");
				exrFrameBuffer.insert("Z.z", Imf::Slice(Imf::FLOAT, (char*)(&depthZPixels32[0]), sizeof(float)*1, sizeof(float)*width));
				break;
			}
		}
	}


//Processing N channel
	if(nPass == miTRUE)
	{
		switch (nFormat)
		{
			case 0:
			{
				exrHeader.channels().insert ("Normal.r", Imf::Channel(Imf::UINT));
				exrHeader.channels().insert ("Normal.g", Imf::Channel(Imf::UINT));
				exrHeader.channels().insert ("Normal.b", Imf::Channel(Imf::UINT));

				fb_nPass = mi_output_image_open(state, miRC_IMAGE_N);

				nRPixels8 = new unsigned int[width*height];
				nGPixels8 = new unsigned int[width*height];
				nBPixels8 = new unsigned int[width*height];

				for (x=0; x<width; x++)
				{
					if (mi_par_aborted())
						break;
					for (y=0; y<height; y++)
					{
						mi_img_get_normal(fb_nPass, &mvector, x, height-y-1);
						nRPixels8[x+y*width] = Imf::floatToUint(mvector.x);
						nGPixels8[x+y*width] = Imf::floatToUint(mvector.y);
						nBPixels8[x+y*width] = Imf::floatToUint(mvector.z);
					}
				}
			    mi_output_image_close(state, miRC_IMAGE_N);
				mi_progress("p_OpenEXR_saver(): Normal...(unint)");
				exrFrameBuffer.insert("Normal.r", Imf::Slice(Imf::UINT, (char*)(&nRPixels8[0]), sizeof(nRPixels8[0])*1, sizeof(nRPixels8[0])*width));
				exrFrameBuffer.insert("Normal.g", Imf::Slice(Imf::UINT, (char*)(&nGPixels8[0]), sizeof(nGPixels8[0])*1, sizeof(nGPixels8[0])*width));
				exrFrameBuffer.insert("Normal.b", Imf::Slice(Imf::UINT, (char*)(&nBPixels8[0]), sizeof(nBPixels8[0])*1, sizeof(nBPixels8[0])*width));
				break;
			}
			case 1:
			{
				exrHeader.channels().insert ("Normal.r", Imf::Channel(Imf::HALF));
				exrHeader.channels().insert ("Normal.g", Imf::Channel(Imf::HALF));
				exrHeader.channels().insert ("Normal.b", Imf::Channel(Imf::HALF));

				fb_nPass = mi_output_image_open(state, miRC_IMAGE_N);

				nRPixels16 = new half[width*height];
				nGPixels16 = new half[width*height];
				nBPixels16 = new half[width*height];

				for (x=0; x<width; x++)
				{
					if (mi_par_aborted())
						break;
					for (y=0; y<height; y++)
					{
						mi_img_get_normal(fb_nPass, &mvector, x, height-y-1);

						nRPixels16[x+y*width] = Imf::floatToHalf(mvector.x);
						nGPixels16[x+y*width] = Imf::floatToHalf(mvector.y);
						nBPixels16[x+y*width] = Imf::floatToHalf(mvector.z);
					}
				}
			    mi_output_image_close(state, miRC_IMAGE_N);
				mi_progress("p_OpenEXR_saver(): Normal...(half)");
				exrFrameBuffer.insert("Normal.r", Imf::Slice(Imf::HALF, (char*)(&nRPixels16[0]), sizeof(nRPixels16[0])*1, sizeof(nRPixels16[0])*width));
				exrFrameBuffer.insert("Normal.g", Imf::Slice(Imf::HALF, (char*)(&nGPixels16[0]), sizeof(nGPixels16[0])*1, sizeof(nGPixels16[0])*width));
				exrFrameBuffer.insert("Normal.b", Imf::Slice(Imf::HALF, (char*)(&nBPixels16[0]), sizeof(nBPixels16[0])*1, sizeof(nBPixels16[0])*width));
				break;
			}
			case 2:
			{
				exrHeader.channels().insert ("Normal.r", Imf::Channel(Imf::FLOAT));
				exrHeader.channels().insert ("Normal.g", Imf::Channel(Imf::FLOAT));
				exrHeader.channels().insert ("Normal.b", Imf::Channel(Imf::FLOAT));

				fb_nPass = mi_output_image_open(state, miRC_IMAGE_N);

				nRPixels32 = new float[width*height];
				nGPixels32 = new float[width*height];
				nBPixels32 = new float[width*height];

				for (x=0; x<width; x++)
				{
					if (mi_par_aborted())
						break;
					for (y=0; y<height; y++)
					{
						mi_img_get_normal(fb_nPass, &mvector, x, height-y-1);
						nRPixels32[x+y*width] = mvector.x;
						nGPixels32[x+y*width] = mvector.y;
						nBPixels32[x+y*width] = mvector.z;
					}
				}
			    mi_output_image_close(state, miRC_IMAGE_N);
				mi_progress("p_OpenEXR_saver(): Normal...(float)");
				exrFrameBuffer.insert("Normal.r",	Imf::Slice(Imf::FLOAT, (char*)(&nRPixels32[0]),	sizeof(nRPixels32[0])*1, sizeof(nRPixels32[0])*width));
				exrFrameBuffer.insert("Normal.g",	Imf::Slice(Imf::FLOAT, (char*)(&nGPixels32[0]),	sizeof(nGPixels32[0])*1, sizeof(nGPixels32[0])*width));
				exrFrameBuffer.insert("Normal.b",	Imf::Slice(Imf::FLOAT, (char*)(&nBPixels32[0]),	sizeof(nBPixels32[0])*1, sizeof(nBPixels32[0])*width));
				break;
			}
		}
	}

//Processing M channel
	if(mPass == miTRUE)
	{
		switch (mFormat)
		{
			case 0:
			{
				exrHeader.channels().insert ("Motion.r", Imf::Channel(Imf::UINT));
				exrHeader.channels().insert ("Motion.g", Imf::Channel(Imf::UINT));
				exrHeader.channels().insert ("Motion.b", Imf::Channel(Imf::UINT));

				fb_mPass = mi_output_image_open(state, miRC_IMAGE_M);

				mRPixels8 = new unsigned int[width*height];
				mGPixels8 = new unsigned int[width*height];
				mBPixels8 = new unsigned int[width*height];

				for (x=0; x<width; x++)
				{
					if (mi_par_aborted())
						break;
					for (y=0; y<height; y++)
					{
						mi_img_get_normal(fb_mPass, &mvector, x, height-y-1);
						mRPixels8[x+y*width] = Imf::floatToUint(mvector.x);
						mGPixels8[x+y*width] = Imf::floatToUint(mvector.y);
						mBPixels8[x+y*width] = Imf::floatToUint(mvector.z);
					}
				}
			    mi_output_image_close(state, miRC_IMAGE_M);
				mi_progress("p_OpenEXR_saver(): Motion...(unint)");
				exrFrameBuffer.insert("Motion.r", Imf::Slice(Imf::UINT, (char*)(&mRPixels8[0]), sizeof(mRPixels8[0])*1, sizeof(mRPixels8[0])*width));
				exrFrameBuffer.insert("Motion.g", Imf::Slice(Imf::UINT, (char*)(&mGPixels8[0]), sizeof(mGPixels8[0])*1, sizeof(mGPixels8[0])*width));
				exrFrameBuffer.insert("Motion.b", Imf::Slice(Imf::UINT, (char*)(&mBPixels8[0]), sizeof(mBPixels8[0])*1, sizeof(mBPixels8[0])*width));
				break;
			}
			case 1:
			{
				exrHeader.channels().insert ("Motion.r", Imf::Channel(Imf::HALF));
				exrHeader.channels().insert ("Motion.g", Imf::Channel(Imf::HALF));
				exrHeader.channels().insert ("Motion.b", Imf::Channel(Imf::HALF));

				fb_mPass = mi_output_image_open(state, miRC_IMAGE_M);

				mRPixels16 = new half[width*height];
				mGPixels16 = new half[width*height];
				mBPixels16 = new half[width*height];

				for (x=0; x<width; x++)
				{
					if (mi_par_aborted())
						break;
					for (y=0; y<height; y++)
					{
						mi_img_get_normal(fb_mPass, &mvector, x, height-y-1);
						mRPixels16[x+y*width] = Imf::floatToHalf(mvector.x);
						mGPixels16[x+y*width] = Imf::floatToHalf(mvector.y);
						mBPixels16[x+y*width] = Imf::floatToHalf(mvector.z);
					}
				}
			    mi_output_image_close(state, miRC_IMAGE_M);
				mi_progress("p_OpenEXR_saver(): Motion...(half)");
				exrFrameBuffer.insert("Motion.r", Imf::Slice(Imf::HALF, (char*)(&mRPixels16[0]), sizeof(mRPixels16[0])*1, sizeof(mRPixels16[0])*width));
				exrFrameBuffer.insert("Motion.g", Imf::Slice(Imf::HALF, (char*)(&mGPixels16[0]), sizeof(mGPixels16[0])*1, sizeof(mGPixels16[0])*width));
				exrFrameBuffer.insert("Motion.b", Imf::Slice(Imf::HALF, (char*)(&mBPixels16[0]), sizeof(mBPixels16[0])*1, sizeof(mBPixels16[0])*width));
				break;
			}

			case 2:
			{
				exrHeader.channels().insert ("Motion.r", Imf::Channel(Imf::FLOAT));
				exrHeader.channels().insert ("Motion.g", Imf::Channel(Imf::FLOAT));
				exrHeader.channels().insert ("Motion.b", Imf::Channel(Imf::FLOAT));

				fb_mPass = mi_output_image_open(state, miRC_IMAGE_M);

				mRPixels32 = new float[width*height];
				mGPixels32 = new float[width*height];
				mBPixels32 = new float[width*height];

				for (x=0; x<width; x++)
				{
					if (mi_par_aborted())
						break;
					for (y=0; y<height; y++)
					{
						mi_img_get_normal(fb_mPass, &mvector, x, height-y-1);
						mRPixels32[x+y*width] = mvector.x;
						mGPixels32[x+y*width] = mvector.y;
						mBPixels32[x+y*width] = mvector.z;
					}
				}
				mi_output_image_close(state, miRC_IMAGE_M);
				mi_progress("p_OpenEXR_saver(): Motion...(float)");
				exrFrameBuffer.insert("Motion.r", Imf::Slice(Imf::FLOAT, (char*)(&mRPixels32[0]), sizeof(float)*1, sizeof(float)*width));
				exrFrameBuffer.insert("Motion.g", Imf::Slice(Imf::FLOAT, (char*)(&mGPixels32[0]), sizeof(float)*1, sizeof(float)*width));
				exrFrameBuffer.insert("Motion.b", Imf::Slice(Imf::FLOAT, (char*)(&mBPixels32[0]), sizeof(float)*1, sizeof(float)*width));
				break;
			}
		}
	}

//Processing TAG channel
	if(tPass == miTRUE)
	{
		switch (tFormat)
		{
			case 0:
			{
				exrHeader.channels().insert ("Label.x", Imf::Channel(Imf::UINT));
			    fb_tPass = mi_output_image_open(state, miRC_IMAGE_TAG);
				tZPixels8 = new unsigned int[width*height];

				for (x=0; x<width; x++)
				{
					if (mi_par_aborted())
						break;
					for (y=0; y<height; y++)
					{
						mi_img_get_label(fb_tPass, &unsint, x, height-y-1);
						tZPixels8[x+y*width] = unsint;
					}
				}
			    mi_output_image_close(state, miRC_IMAGE_TAG);
				mi_progress("p_OpenEXR_saver(): Label...(unint)");
				exrFrameBuffer.insert("Label.x", Imf::Slice(Imf::UINT, (char*)(&tZPixels8[0]), sizeof(tZPixels8[0])*1, sizeof(tZPixels8[0])*width));
				break;
			}

			case 1:
			{
				exrHeader.channels().insert ("Label.x", Imf::Channel(Imf::HALF));
			    fb_tPass = mi_output_image_open(state, miRC_IMAGE_TAG);
				tZPixels16 = new half[width*height];

				for (x=0; x<width; x++)
				{
					if (mi_par_aborted())
						break;
					for (y=0; y<height; y++)
					{
						mi_img_get_label(fb_tPass, &unsint, x, height-y-1);
//						tZPixels16[x+y*width] = Imf::floatToHalf(scalar);
						tZPixels16[x+y*width] = Imf::uintToHalf(unsint);
					}
				}
				mi_output_image_close(state, miRC_IMAGE_TAG);
				mi_progress("p_OpenEXR_saver(): Label...(half)");
				exrFrameBuffer.insert("Label.x", Imf::Slice(Imf::HALF, (char*)(&tZPixels16[0]), sizeof(tZPixels16[0])*1, sizeof(tZPixels16[0])*width));
				break;
			}
			case 2:
			{
				exrHeader.channels().insert ("Label.x", Imf::Channel(Imf::FLOAT));
			    fb_tPass = mi_output_image_open(state, miRC_IMAGE_TAG);
				tZPixels32 = new float[width*height];

				for (x=0; x<width; x++)
				{
					if (mi_par_aborted())
						break;
					for (y=0; y<height; y++)
					{
						mi_img_get_label(fb_tPass, &unsint, x, height-y-1);
						tZPixels32[x+y*width] = (float)unsint;
//						tZPixels32[x+y*width] = Imf::uintToFloat(unsint);
					}
				}
				mi_output_image_close(state, miRC_IMAGE_TAG);
				mi_progress("p_OpenEXR_saver(): Label...(float)");
				exrFrameBuffer.insert("Label.x", Imf::Slice(Imf::FLOAT, (char*)(&tZPixels32[0]), sizeof(tZPixels32[0])*1, sizeof(tZPixels32[0])*width));
				break;
			}
		}
	}

//Processing COVERAGE channel
	if(cPass == miTRUE)
	{
		switch (cFormat)
		{
			case 0:
			{
				exrHeader.channels().insert ("Coverage.x", Imf::Channel(Imf::UINT));
			    fb_cPass = mi_output_image_open(state, miRC_IMAGE_COVERAGE);
				cZPixels8 = new unsigned int[width*height];

				for (x=0; x<width; x++)
				{
					if (mi_par_aborted())
						break;
					for (y=0; y<height; y++)
					{
						mi_img_get_depth(fb_cPass, &scalar, x, height-y-1);
						cZPixels8[x+y*width] = Imf::floatToUint(scalar);
					}
				}
			    mi_output_image_close(state, miRC_IMAGE_COVERAGE);
				mi_progress("p_OpenEXR_saver(): Coverage...(unint)");
				exrFrameBuffer.insert("Coverage.x", Imf::Slice(Imf::UINT, (char*)(&cZPixels8[0]), sizeof(unsigned int)*1, sizeof(unsigned int)*width));
				break;
			}

			case 1:
			{
				exrHeader.channels().insert ("Coverage.x", Imf::Channel(Imf::HALF));
			    fb_cPass = mi_output_image_open(state, miRC_IMAGE_COVERAGE);
				cZPixels16 = new half[width*height];

				for (x=0; x<width; x++)
				{
					if (mi_par_aborted())
						break;
					for (y=0; y<height; y++)
					{
						mi_img_get_depth(fb_cPass, &scalar, x, height-y-1);
						cZPixels16[x+y*width] = Imf::floatToHalf(scalar);
					}
				}
				mi_output_image_close(state, miRC_IMAGE_COVERAGE);
				mi_progress("p_OpenEXR_saver(): Coverage...(half)");
				exrFrameBuffer.insert("Coverage.x", Imf::Slice(Imf::HALF, (char*)(&cZPixels16[0]), sizeof(half)*1, sizeof(half)*width));
				break;
			}
			case 2:
			{
				exrHeader.channels().insert ("Coverage.x", Imf::Channel(Imf::FLOAT));
			    fb_cPass = mi_output_image_open(state, miRC_IMAGE_COVERAGE);
				cZPixels32 = new float[width*height];

				for (x=0; x<width; x++)
				{
					if (mi_par_aborted())
						break;
					for (y=0; y<height; y++)
					{
						mi_img_get_depth(fb_cPass, &scalar, x, height-y-1);
						cZPixels32[x+y*width] = scalar;
					}
				}
				mi_output_image_close(state, miRC_IMAGE_COVERAGE);
				mi_progress("p_OpenEXR_saver(): Coverage...(float)");
				exrFrameBuffer.insert("Coverage.x", Imf::Slice(Imf::FLOAT, (char*)(&cZPixels32[0]), sizeof(float)*1, sizeof(float)*width));
				break;
			}
		}
	}

	pass_pt Shading_pass_data;
	pass_pt Color_pass_data;
	pass_pt Diffuse_pass_data;
	pass_pt Shadow_pass_data;
	pass_pt Specular_pass_data;
	pass_pt SpecularShd_pass_data;
	pass_pt DiffuseOcclusion_pass_data;
	pass_pt IndirectIllum_pass_data;
	pass_pt Ambient_pass_data;
	pass_pt AmbOcclusion_pass_data;
	pass_pt Reflect_pass_data;
	pass_pt Refract_pass_data;
	pass_pt Incandescence_pass_data;
	pass_pt Depth_pass_data;
	pass_pt BentNormal_pass_data;
	pass_pt SpecialC1_pass_data;
	pass_pt SpecialC2_pass_data;
	pass_pt SpecialC3_pass_data;
	pass_pt SpecialC4_pass_data;
	pass_pt SpecialC5_pass_data;
	pass_pt SpecialC6_pass_data;
	pass_pt SpecialC7_pass_data;
	pass_pt SpecialC8_pass_data;
	pass_pt SpecialC9_pass_data;
	pass_pt SpecialC10_pass_data;

	if(shadingPass == miTRUE)
	{
		if(save_pass(state, shadingBuffer, shadingFormat, width, height, "p_MegaTK_shading", &exrHeader, &exrFrameBuffer, &Shading_pass_data) != 1)
		{				
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_shading pass channels");
			shadingPass = miFALSE;
		}
	}
	if(colorPass == miTRUE)
	{
		if(save_pass(state, colorBuffer, colorFormat, width, height, "p_MegaTK_color", &exrHeader, &exrFrameBuffer, &Color_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_color pass channels");
			colorPass = miFALSE;
		}
	}
	if(diffusePass == miTRUE)
	{
		if(save_pass(state, diffuseBuffer, diffuseFormat, width, height, "p_MegaTK_diffuse", &exrHeader, &exrFrameBuffer, &Diffuse_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_diffuse pass channels");
			diffusePass = miFALSE;
		}
	}
	if(diffuseShadowPass == miTRUE)
	{
		if(save_pass(state, diffuseShadowBuffer, diffuseShadowFormat, width, height, "p_MegaTK_shadow", &exrHeader, &exrFrameBuffer, &Shadow_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_shadow pass channels");
			diffuseShadowPass = miFALSE;
		}
	}
	if(specularPass == miTRUE)
	{
		if(save_pass(state, specularBuffer, specularFormat, width, height, "p_MegaTK_specular", &exrHeader, &exrFrameBuffer, &Specular_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_specular pass channels");
			specularPass = miFALSE;
		}
	}
	if(specularShadowPass == miTRUE)
	{
		if(save_pass(state, specularShadowBuffer, specularShadowFormat, width, height, "p_MegaTK_spec_shd", &exrHeader, &exrFrameBuffer, &SpecularShd_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_spec_shd pass channels");
			specularShadowPass = miFALSE;
		}
	}
	if(diffuseOcclusionPass == miTRUE)
	{
		if(save_pass(state, diffuseOcclusionBuffer, diffuseOcclusionFormat, width, height, "p_MegaTK_diff_occlus", &exrHeader, &exrFrameBuffer, &DiffuseOcclusion_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_diff_occlus pass channels");
			diffuseOcclusionPass = miFALSE;
		}
	}
	if(indirectIllumPass == miTRUE)
	{
		if(save_pass(state, indirectIllumBuffer, indirectIllumFormat, width, height, "p_MegaTK_indirect", &exrHeader, &exrFrameBuffer, &IndirectIllum_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_indirect pass channels");
			indirectIllumPass = miFALSE;
		}
	}
	if(ambientPass == miTRUE)
	{
		if(save_pass(state, ambientBuffer, ambientFormat, width, height, "p_MegaTK_ambient", &exrHeader, &exrFrameBuffer, &Ambient_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_ambient pass channels");
			ambientPass = miFALSE;
		}
	}
	if(ambOcclusionPass == miTRUE)
	{
		if(save_pass(state, ambOcclusionBuffer, ambOcclusionFormat, width, height, "p_MegaTK_amb_occlus", &exrHeader, &exrFrameBuffer, &AmbOcclusion_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_amb_occlus pass channels");
			ambOcclusionPass = miFALSE;
		}
	}
	if(reflectPass == miTRUE)
	{
		if(save_pass(state, reflectBuffer, reflectFormat, width, height, "p_MegaTK_reflect", &exrHeader, &exrFrameBuffer, &Reflect_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_reflect pass channels");
			reflectPass = miFALSE;
		}
	}
	if(refractPass == miTRUE)
	{
		if(save_pass(state, refractBuffer, refractFormat, width, height, "p_MegaTK_refract", &exrHeader, &exrFrameBuffer, &Refract_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_refract pass channels");
			refractPass = miFALSE;
		}
	}
	if(incandescencePass == miTRUE)
	{
		if(save_pass(state, incandescenceBuffer, incandescenceFormat, width, height, "p_MegaTK_incand", &exrHeader, &exrFrameBuffer, &Incandescence_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_incand pass channels");
			incandescencePass = miFALSE;
		}
	}
	if(depthPass == miTRUE)
	{
		if(save_pass(state, depthBuffer, depthFormat, width, height, "p_MegaTK_z", &exrHeader, &exrFrameBuffer, &Depth_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_z pass channels");
			depthPass = miFALSE;
		}
	}
	if(bentNormalPass == miTRUE)
	{
		if(save_pass(state, bentNormalBuffer, bentNormalFormat, width, height, "p_MegaTK_bent", &exrHeader, &exrFrameBuffer, &BentNormal_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_bent pass channels");
			bentNormalPass = miFALSE;
		}
	}
	if(specialC1Pass == miTRUE)
	{
		if(save_pass(state, specialC1Buffer, specialC1Format, width, height, "p_MegaTK_specialC1", &exrHeader, &exrFrameBuffer, &SpecialC1_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_specialC1 pass channels");
			specialC1Pass = miFALSE;
		}
	}
	if(specialC2Pass == miTRUE)
	{
		if(save_pass(state, specialC2Buffer, specialC2Format, width, height, "p_MegaTK_specialC2", &exrHeader, &exrFrameBuffer, &SpecialC2_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_specialC2 pass channels");
			specialC2Pass = miFALSE;
		}
	}
	if(specialC3Pass == miTRUE)
	{
		if(save_pass(state, specialC3Buffer, specialC3Format, width, height, "p_MegaTK_specialC3", &exrHeader, &exrFrameBuffer, &SpecialC3_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_specialC3 pass channels");
			specialC3Pass = miFALSE;
		}
	}
	if(specialC4Pass == miTRUE)
	{
		if(save_pass(state, specialC4Buffer, specialC4Format, width, height, "p_MegaTK_specialC4", &exrHeader, &exrFrameBuffer, &SpecialC4_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_specialC4 pass channels");
			specialC4Pass = miFALSE;
		}
	}
	if(specialC5Pass == miTRUE)
	{
		if(save_pass(state, specialC5Buffer, specialC5Format, width, height, "p_MegaTK_specialC5", &exrHeader, &exrFrameBuffer, &SpecialC5_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_specialC5 pass channels");
			specialC5Pass = miFALSE;
		}
	}
	if(specialC6Pass == miTRUE)
	{
		if(save_pass(state, specialC6Buffer, specialC6Format, width, height, "p_MegaTK_specialC6", &exrHeader, &exrFrameBuffer, &SpecialC6_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_specialC6 pass channels");
			specialC6Pass = miFALSE;
		}
	}
	if(specialC7Pass == miTRUE)
	{
		if(save_pass(state, specialC7Buffer, specialC7Format, width, height, "p_MegaTK_specialC7", &exrHeader, &exrFrameBuffer, &SpecialC7_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_specialC7 pass channels");
			specialC7Pass = miFALSE;
		}
	}
	if(specialC8Pass == miTRUE)
	{
		if(save_pass(state, specialC8Buffer, specialC8Format, width, height, "p_MegaTK_specialC8", &exrHeader, &exrFrameBuffer, &SpecialC8_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_specialC8 pass channels");
			specialC8Pass = miFALSE;
		}
	}
	if(specialC9Pass == miTRUE)
	{
		if(save_pass(state, specialC9Buffer, specialC9Format, width, height, "p_MegaTK_specialC9", &exrHeader, &exrFrameBuffer, &SpecialC9_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_specialC9 pass channels");
			specialC9Pass = miFALSE;
		}
	}
	if(specialC10Pass == miTRUE)
	{
		if(save_pass(state, specialC10Buffer, specialC10Format, width, height, "p_MegaTK_specialC10", &exrHeader, &exrFrameBuffer, &SpecialC10_pass_data) != 1)
		{
			mi_warning("p_OpenEXR_saver(): failed to create p_MegaTK_specialC10 pass channels");
			specialC10Pass = miFALSE;
		}
	}

	//add meta data

	exrHeader.insert ("comments", Imf::StringAttribute("Created using mental ray p_OpenEXR_saver shader"));
	exrHeader.insert ("aperture", Imf::FloatAttribute(state->camera->aperture));
	exrHeader.insert ("focal", Imf::FloatAttribute(state->camera->focal * ITM));

	float angleOfView = (float)(RTA * atan((state->camera->aperture * ITM / 2) / (state->camera->focal * ITM)) * 2);
	exrHeader.insert ("angleOfView", Imf::FloatAttribute(angleOfView));

	exrHeader.insert ("clipNear", Imf::FloatAttribute(state->camera->clip.min));
	exrHeader.insert ("clipFar", Imf::FloatAttribute(state->camera->clip.max));

	switch(compression)
	{
		case 0:
			exrHeader.insert ("compress", Imf::StringAttribute("none"));
			break;
		case 1:
			exrHeader.insert ("compress", Imf::StringAttribute("rle"));
			break;
		case 2:
			exrHeader.insert ("compress", Imf::StringAttribute("zip1"));
			break;
		case 3:
			exrHeader.insert ("compress", Imf::StringAttribute("zip16"));
			break;
		case 4:
			exrHeader.insert ("compress", Imf::StringAttribute("piz"));
			break;
		case 5:
			exrHeader.insert ("compress", Imf::StringAttribute("pxr24"));
			break;
		case 6:
			exrHeader.insert ("compress", Imf::StringAttribute("b44"));
			break;
		case 7:
			exrHeader.insert ("compress", Imf::StringAttribute("b44a"));
			break;
		default:
			exrHeader.insert ("compress", Imf::StringAttribute("none"));
			break;
	}

	miInstance *inst = (miInstance *)mi_db_access(state->camera_inst);
	miTransform cameraTF = inst->tf;
	miMatrix cameraMTF;	
	for(int i=0; i<16; i++)
		cameraMTF[i] = inst->motion_transform[i];

	mi_db_unpin(state->camera_inst);

	Imath::M44f cameraTransform(cameraTF.local_to_global[0], cameraTF.local_to_global[1], cameraTF.local_to_global[2], cameraTF.local_to_global[3],
		cameraTF.local_to_global[4], cameraTF.local_to_global[5], cameraTF.local_to_global[6], cameraTF.local_to_global[7],
		cameraTF.local_to_global[8], cameraTF.local_to_global[9], cameraTF.local_to_global[10], cameraTF.local_to_global[11],
		cameraTF.local_to_global[12], cameraTF.local_to_global[13], cameraTF.local_to_global[14], cameraTF.local_to_global[15]);

	Imath::M44f cameraMotion(cameraMTF[0], cameraMTF[1], cameraMTF[2], cameraMTF[3],
		cameraMTF[4], cameraMTF[5], cameraMTF[6], cameraMTF[7],
		cameraMTF[8], cameraMTF[9], cameraMTF[10], cameraMTF[11],
		cameraMTF[12], cameraMTF[13], cameraMTF[14], cameraMTF[15]);

	exrHeader.insert("cameraTransform", Imf::M44fAttribute(cameraTransform));
	exrHeader.insert("cameraMotion", Imf::M44fAttribute(cameraMotion));

	try
	{
		Imf::OutputFile exrFile (fileName, exrHeader);		//create file
		exrFile.setFrameBuffer(exrFrameBuffer);
		exrFile.writePixels(height);					//write file
	}
	catch(...)
	{
		mi_error("p_OpenEXR_saver(): Error writing exr file. Please check your permission or free space.");
		return miFALSE;
	}

	//cleanup
	if(rgbaPass == miTRUE)
	{
		switch (rgbaFormat)
		{
			case 0:
				delete rgbaRPixels8;
				delete rgbaGPixels8;
				delete rgbaBPixels8;
				delete rgbaAPixels8;
				break;
			case 1:
				delete rgbaRPixels16;
				delete rgbaGPixels16;
				delete rgbaBPixels16;
				delete rgbaAPixels16;
				break;
			case 2:
				delete rgbaRPixels32;
				delete rgbaGPixels32;
				delete rgbaBPixels32;
				delete rgbaAPixels32;
				break;
		}
	}
	if(zPass == miTRUE)
	{
		switch (zFormat)
		{
			case 0:
				delete depthZPixels8;
				break;
			case 1:
				delete depthZPixels16;
				break;
			case 2:
				delete depthZPixels32;
				break;
		}
	}
	if(nPass == miTRUE)
	{
		switch (nFormat)
		{
			case 0:
				delete nRPixels8;
				delete nGPixels8;
				delete nBPixels8;
				break;
			case 1:
				delete nRPixels16;
				delete nGPixels16;
				delete nBPixels16;
				break;

			case 2:
				delete nRPixels32;
				delete nGPixels32;
				delete nBPixels32;
				break;
		}
	}
	if(mPass == miTRUE)
	{
		switch (mFormat)
		{
			case 0:
				delete mRPixels8;
				delete  mGPixels8;
				delete  mBPixels8;
				break;

			case 1:
				delete  mRPixels16;
				delete  mGPixels16;
				delete  mBPixels16;
				break;

			case 2:
				delete  mRPixels32;
				delete  mGPixels32;
				delete  mBPixels32;
				break;
		}
	}

	if(tPass == miTRUE)
	{
		switch (tFormat)
		{
			case 0:
				delete  tZPixels8;
				break;
			case 1:
				delete  tZPixels16;
				break;
			case 2:
				delete  tZPixels32;
				break;
		}
	}
	if(cPass == miTRUE)
	{
		switch (cFormat)
		{
			case 0:
				delete  cZPixels8;
				break;
			case 1:
				delete  cZPixels16;
				break;
			case 2:
				delete  cZPixels32;
				break;
		}
	}


	if(shadingPass == miTRUE)
	{
		switch(shadingFormat)
		{
			case 0:
				mi_mem_release(Shading_pass_data.unintR);
				mi_mem_release(Shading_pass_data.unintG);
				mi_mem_release(Shading_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(Shading_pass_data.halfR);
				mi_mem_release(Shading_pass_data.halfG);
				mi_mem_release(Shading_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(Shading_pass_data.floatR);
				mi_mem_release(Shading_pass_data.floatG);
				mi_mem_release(Shading_pass_data.floatB);
				break;
		}
	}
	if(colorPass == miTRUE)
	{
		switch(colorFormat)
		{
			case 0:
				mi_mem_release(Color_pass_data.unintR);
				mi_mem_release(Color_pass_data.unintG);
				mi_mem_release(Color_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(Color_pass_data.halfR);
				mi_mem_release(Color_pass_data.halfG);
				mi_mem_release(Color_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(Color_pass_data.floatR);
				mi_mem_release(Color_pass_data.floatG);
				mi_mem_release(Color_pass_data.floatB);
				break;
		}
	}
	if(diffusePass == miTRUE)
	{
		switch(diffuseFormat)
		{
			case 0:
				mi_mem_release(Diffuse_pass_data.unintR);
				mi_mem_release(Diffuse_pass_data.unintG);
				mi_mem_release(Diffuse_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(Diffuse_pass_data.halfR);
				mi_mem_release(Diffuse_pass_data.halfG);
				mi_mem_release(Diffuse_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(Diffuse_pass_data.floatR);
				mi_mem_release(Diffuse_pass_data.floatG);
				mi_mem_release(Diffuse_pass_data.floatB);
				break;
		}
	}
	if(diffuseShadowPass == miTRUE)
	{
		switch(diffuseShadowFormat)
		{
			case 0:
				mi_mem_release(Shadow_pass_data.unintR);
				mi_mem_release(Shadow_pass_data.unintG);
				mi_mem_release(Shadow_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(Shadow_pass_data.halfR);
				mi_mem_release(Shadow_pass_data.halfG);
				mi_mem_release(Shadow_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(Shadow_pass_data.floatR);
				mi_mem_release(Shadow_pass_data.floatG);
				mi_mem_release(Shadow_pass_data.floatB);
				break;
		}
	}
	if(specularPass == miTRUE)
	{
		switch(specularFormat)
		{
			case 0:
				mi_mem_release(Specular_pass_data.unintR);
				mi_mem_release(Specular_pass_data.unintG);
				mi_mem_release(Specular_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(Specular_pass_data.halfR);
				mi_mem_release(Specular_pass_data.halfG);
				mi_mem_release(Specular_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(Specular_pass_data.floatR);
				mi_mem_release(Specular_pass_data.floatG);
				mi_mem_release(Specular_pass_data.floatB);
				break;
		}
	}
	if(specularShadowPass == miTRUE)
	{
		switch(specularShadowFormat)
		{
			case 0:
				mi_mem_release(SpecularShd_pass_data.unintR);
				mi_mem_release(SpecularShd_pass_data.unintG);
				mi_mem_release(SpecularShd_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(SpecularShd_pass_data.halfR);
				mi_mem_release(SpecularShd_pass_data.halfG);
				mi_mem_release(SpecularShd_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(SpecularShd_pass_data.floatR);
				mi_mem_release(SpecularShd_pass_data.floatG);
				mi_mem_release(SpecularShd_pass_data.floatB);
				break;
		}
	}
	if(diffuseOcclusionPass == miTRUE)
	{
		switch(diffuseOcclusionFormat)
		{
			case 0:
				mi_mem_release(DiffuseOcclusion_pass_data.unintR);
				mi_mem_release(DiffuseOcclusion_pass_data.unintG);
				mi_mem_release(DiffuseOcclusion_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(DiffuseOcclusion_pass_data.halfR);
				mi_mem_release(DiffuseOcclusion_pass_data.halfG);
				mi_mem_release(DiffuseOcclusion_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(DiffuseOcclusion_pass_data.floatR);
				mi_mem_release(DiffuseOcclusion_pass_data.floatG);
				mi_mem_release(DiffuseOcclusion_pass_data.floatB);
				break;
		}
	}
	if(indirectIllumPass == miTRUE)
	{
		switch(indirectIllumFormat)
		{
			case 0:
				mi_mem_release(IndirectIllum_pass_data.unintR);
				mi_mem_release(IndirectIllum_pass_data.unintG);
				mi_mem_release(IndirectIllum_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(IndirectIllum_pass_data.halfR);
				mi_mem_release(IndirectIllum_pass_data.halfG);
				mi_mem_release(IndirectIllum_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(IndirectIllum_pass_data.floatR);
				mi_mem_release(IndirectIllum_pass_data.floatG);
				mi_mem_release(IndirectIllum_pass_data.floatB);
				break;
		}
	}
	if(ambientPass == miTRUE)
	{
		switch(ambientFormat)
		{
			case 0:
				mi_mem_release(Ambient_pass_data.unintR);
				mi_mem_release(Ambient_pass_data.unintG);
				mi_mem_release(Ambient_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(Ambient_pass_data.halfR);
				mi_mem_release(Ambient_pass_data.halfG);
				mi_mem_release(Ambient_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(Ambient_pass_data.floatR);
				mi_mem_release(Ambient_pass_data.floatG);
				mi_mem_release(Ambient_pass_data.floatB);
				break;
		}
	}
	if(ambOcclusionPass == miTRUE)
	{
		switch(ambOcclusionFormat)
		{
			case 0:
				mi_mem_release(AmbOcclusion_pass_data.unintR);
				mi_mem_release(AmbOcclusion_pass_data.unintG);
				mi_mem_release(AmbOcclusion_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(AmbOcclusion_pass_data.halfR);
				mi_mem_release(AmbOcclusion_pass_data.halfG);
				mi_mem_release(AmbOcclusion_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(AmbOcclusion_pass_data.floatR);
				mi_mem_release(AmbOcclusion_pass_data.floatG);
				mi_mem_release(AmbOcclusion_pass_data.floatB);
				break;
		}
	}
	if(reflectPass == miTRUE)
	{
		switch(reflectFormat)
		{
			case 0:
				mi_mem_release(Reflect_pass_data.unintR);
				mi_mem_release(Reflect_pass_data.unintG);
				mi_mem_release(Reflect_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(Reflect_pass_data.halfR);
				mi_mem_release(Reflect_pass_data.halfG);
				mi_mem_release(Reflect_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(Reflect_pass_data.floatR);
				mi_mem_release(Reflect_pass_data.floatG);
				mi_mem_release(Reflect_pass_data.floatB);
				break;
		}
	}
	if(refractPass == miTRUE)
	{
		switch(refractFormat)
		{
			case 0:
				mi_mem_release(Refract_pass_data.unintR);
				mi_mem_release(Refract_pass_data.unintG);
				mi_mem_release(Refract_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(Refract_pass_data.halfR);
				mi_mem_release(Refract_pass_data.halfG);
				mi_mem_release(Refract_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(Refract_pass_data.floatR);
				mi_mem_release(Refract_pass_data.floatG);
				mi_mem_release(Refract_pass_data.floatB);
				break;
		}
	}
	if(incandescencePass == miTRUE)
	{
		switch(incandescenceFormat)
		{
			case 0:
				mi_mem_release(Incandescence_pass_data.unintR);
				mi_mem_release(Incandescence_pass_data.unintG);
				mi_mem_release(Incandescence_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(Incandescence_pass_data.halfR);
				mi_mem_release(Incandescence_pass_data.halfG);
				mi_mem_release(Incandescence_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(Incandescence_pass_data.floatR);
				mi_mem_release(Incandescence_pass_data.floatG);
				mi_mem_release(Incandescence_pass_data.floatB);
				break;
		}
	}
	if(depthPass == miTRUE)
	{
		switch(depthFormat)
		{
			case 0:
				mi_mem_release(Depth_pass_data.unintR);
				mi_mem_release(Depth_pass_data.unintG);
				mi_mem_release(Depth_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(Depth_pass_data.halfR);
				mi_mem_release(Depth_pass_data.halfG);
				mi_mem_release(Depth_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(Depth_pass_data.floatR);
				mi_mem_release(Depth_pass_data.floatG);
				mi_mem_release(Depth_pass_data.floatB);
				break;
		}
	}
	if(bentNormalPass == miTRUE)
	{
		switch(bentNormalFormat)
		{
			case 0:
				mi_mem_release(BentNormal_pass_data.unintR);
				mi_mem_release(BentNormal_pass_data.unintG);
				mi_mem_release(BentNormal_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(BentNormal_pass_data.halfR);
				mi_mem_release(BentNormal_pass_data.halfG);
				mi_mem_release(BentNormal_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(BentNormal_pass_data.floatR);
				mi_mem_release(BentNormal_pass_data.floatG);
				mi_mem_release(BentNormal_pass_data.floatB);
				break;
		}
	}
	if(specialC1Pass == miTRUE)
	{
		switch(specialC1Format)
		{
			case 0:
				mi_mem_release(SpecialC1_pass_data.unintR);
				mi_mem_release(SpecialC1_pass_data.unintG);
				mi_mem_release(SpecialC1_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(SpecialC1_pass_data.halfR);
				mi_mem_release(SpecialC1_pass_data.halfG);
				mi_mem_release(SpecialC1_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(SpecialC1_pass_data.floatR);
				mi_mem_release(SpecialC1_pass_data.floatG);
				mi_mem_release(SpecialC1_pass_data.floatB);
				break;
		}
	}
	if(specialC2Pass == miTRUE)
	{
		switch(specialC2Format)
		{
			case 0:
				mi_mem_release(SpecialC2_pass_data.unintR);
				mi_mem_release(SpecialC2_pass_data.unintG);
				mi_mem_release(SpecialC2_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(SpecialC2_pass_data.halfR);
				mi_mem_release(SpecialC2_pass_data.halfG);
				mi_mem_release(SpecialC2_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(SpecialC2_pass_data.floatR);
				mi_mem_release(SpecialC2_pass_data.floatG);
				mi_mem_release(SpecialC2_pass_data.floatB);
				break;
		}
	}
	if(specialC3Pass == miTRUE)
	{
		switch(specialC3Format)
		{
			case 0:
				mi_mem_release(SpecialC3_pass_data.unintR);
				mi_mem_release(SpecialC3_pass_data.unintG);
				mi_mem_release(SpecialC3_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(SpecialC3_pass_data.halfR);
				mi_mem_release(SpecialC3_pass_data.halfG);
				mi_mem_release(SpecialC3_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(SpecialC3_pass_data.floatR);
				mi_mem_release(SpecialC3_pass_data.floatG);
				mi_mem_release(SpecialC3_pass_data.floatB);
				break;
		}
	}
	if(specialC4Pass == miTRUE)
	{
		switch(specialC4Format)
		{
			case 0:
				mi_mem_release(SpecialC4_pass_data.unintR);
				mi_mem_release(SpecialC4_pass_data.unintG);
				mi_mem_release(SpecialC4_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(SpecialC4_pass_data.halfR);
				mi_mem_release(SpecialC4_pass_data.halfG);
				mi_mem_release(SpecialC4_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(SpecialC4_pass_data.floatR);
				mi_mem_release(SpecialC4_pass_data.floatG);
				mi_mem_release(SpecialC4_pass_data.floatB);
				break;
		}
	}
	if(specialC5Pass == miTRUE)
	{
		switch(specialC5Format)
		{
			case 0:
				mi_mem_release(SpecialC5_pass_data.unintR);
				mi_mem_release(SpecialC5_pass_data.unintG);
				mi_mem_release(SpecialC5_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(SpecialC5_pass_data.halfR);
				mi_mem_release(SpecialC5_pass_data.halfG);
				mi_mem_release(SpecialC5_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(SpecialC5_pass_data.floatR);
				mi_mem_release(SpecialC5_pass_data.floatG);
				mi_mem_release(SpecialC5_pass_data.floatB);
				break;
		}
	}
	if(specialC6Pass == miTRUE)
	{
		switch(specialC6Format)
		{
			case 0:
				mi_mem_release(SpecialC6_pass_data.unintR);
				mi_mem_release(SpecialC6_pass_data.unintG);
				mi_mem_release(SpecialC6_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(SpecialC6_pass_data.halfR);
				mi_mem_release(SpecialC6_pass_data.halfG);
				mi_mem_release(SpecialC6_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(SpecialC6_pass_data.floatR);
				mi_mem_release(SpecialC6_pass_data.floatG);
				mi_mem_release(SpecialC6_pass_data.floatB);
				break;
		}
	}
	if(specialC7Pass == miTRUE)
	{
		switch(specialC7Format)
		{
			case 0:
				mi_mem_release(SpecialC7_pass_data.unintR);
				mi_mem_release(SpecialC7_pass_data.unintG);
				mi_mem_release(SpecialC7_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(SpecialC7_pass_data.halfR);
				mi_mem_release(SpecialC7_pass_data.halfG);
				mi_mem_release(SpecialC7_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(SpecialC7_pass_data.floatR);
				mi_mem_release(SpecialC7_pass_data.floatG);
				mi_mem_release(SpecialC7_pass_data.floatB);
				break;
		}
	}
	if(specialC8Pass == miTRUE)
	{
		switch(specialC8Format)
		{
			case 0:
				mi_mem_release(SpecialC8_pass_data.unintR);
				mi_mem_release(SpecialC8_pass_data.unintG);
				mi_mem_release(SpecialC8_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(SpecialC8_pass_data.halfR);
				mi_mem_release(SpecialC8_pass_data.halfG);
				mi_mem_release(SpecialC8_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(SpecialC8_pass_data.floatR);
				mi_mem_release(SpecialC8_pass_data.floatG);
				mi_mem_release(SpecialC8_pass_data.floatB);
				break;
		}
	}
	if(specialC9Pass == miTRUE)
	{
		switch(specialC9Format)
		{
			case 0:
				mi_mem_release(SpecialC9_pass_data.unintR);
				mi_mem_release(SpecialC9_pass_data.unintG);
				mi_mem_release(SpecialC9_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(SpecialC9_pass_data.halfR);
				mi_mem_release(SpecialC9_pass_data.halfG);
				mi_mem_release(SpecialC9_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(SpecialC9_pass_data.floatR);
				mi_mem_release(SpecialC9_pass_data.floatG);
				mi_mem_release(SpecialC9_pass_data.floatB);
				break;
		}
	}
	if(specialC10Pass == miTRUE)
	{
		switch(specialC10Format)
		{
			case 0:
				mi_mem_release(SpecialC10_pass_data.unintR);
				mi_mem_release(SpecialC10_pass_data.unintG);
				mi_mem_release(SpecialC10_pass_data.unintB);
				break;
			case 1:
				mi_mem_release(SpecialC10_pass_data.halfR);
				mi_mem_release(SpecialC10_pass_data.halfG);
				mi_mem_release(SpecialC10_pass_data.halfB);
				break;
			case 2:
				mi_mem_release(SpecialC10_pass_data.floatR);
				mi_mem_release(SpecialC10_pass_data.floatG);
				mi_mem_release(SpecialC10_pass_data.floatB);
				break;
		}
	}

	mi_progress("p_OpenEXR_saver(): File saved!");

	return(miTRUE);
}
}
