//  Copyright (c) 2005, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "shader.h"
#include "geoshader.h"

typedef struct
{
	int i_objects;
	int n_objects;
	miTag objects[1];
	miVector coord;
	int i_textures;
	int n_textures;
	miTag textures[1];
}p_bake_read_t;

extern "C"
{
DLLEXPORT int p_bake_read_version(void) {return(2);}

DLLEXPORT miBoolean p_bake_read(miColor *result, miState *state, p_bake_read_t *paras)
{
	if(state->type == miRAY_SHADOW || state->type == miRAY_DISPLACE)
		return miFALSE;

//Get name of object (trasform)
	assert(state->instance);
    miInstance *inst = (miInstance *)mi_db_access(state->instance);
    const char *obj_name = inst->parent ? mi_api_tag_lookup(inst->parent) : 0;
//	const char *obj_name = inst->item ? mi_api_tag_lookup(inst->item) : 0;
    mi_db_unpin(state->instance);

//	mi_warning("%s", obj_name);


//Normalize UV to 0-1
	miVector *coord = mi_eval_vector(&paras->coord);
	while(coord->x > 1)
		coord->x = coord->x -1;
	while(coord->x < 0)
		coord->x = coord->x +1;
	while(coord->y > 1)
		coord->y = coord->y -1;
	while(coord->y < 0)
		coord->y = coord->y +1;


	miTag *textures;
	int i_t;
	int n_t;
	i_t = *mi_eval_integer(&paras->i_textures);
	n_t = *mi_eval_integer(&paras->n_textures);
	textures = mi_eval_tag(paras->textures) + i_t;

	miTag *objects;
	int i_o;
	int n_o;
	i_o = *mi_eval_integer(&paras->i_objects);
	n_o = *mi_eval_integer(&paras->n_objects);
	objects = mi_eval_tag(paras->objects) + i_o;

	for (int i = 0; i < i_t; i++)
	{
		if (!objects[i])
		{
			result->b = 1;
			result->g = result->r = result->a = 0;
			return(miTRUE);
		}
		char *filename = (char*)mi_db_access(objects[i]);
		mi_db_unpin(objects[i]);
//		mi_warning("%s", filename);
		if(strcmp(obj_name, filename) == 0)
		{
 			if (textures[i] && coord->x >= 0 && coord->x < 1 && coord->y >= 0 && coord->y < 1 && mi_lookup_color_texture(result, state, textures[i], coord))
				return(miTRUE);
			else
			{
				result->r = result->g = result->b = result->a = 1;
				return(miTRUE);
			}
		}
	}

//Error
	result->r = 1;
	result->g = result->b = result->a = 0;
	return(miTRUE);
}
}
