//  Copyright (c) 2005, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <stdio.h>
#include <math.h>
#include "shader.h"

typedef struct
{
	miColor render;
	miBoolean enable_fb0;
	miColor buffer_0;
	miBoolean enable_fb1;
	miColor buffer_1;
	miBoolean enable_fb2;
	miColor buffer_2;
	miBoolean enable_fb3;
	miColor buffer_3;
	miBoolean enable_fb4;
	miColor buffer_4;
	miBoolean enable_fb5;
	miColor buffer_5;
	miBoolean enable_fb6;
	miColor buffer_6;
	miBoolean enable_fb7;
	miColor buffer_7;
	miBoolean restore_normal;
}p_buffer_writer_t;

extern "C"
{
DLLEXPORT int p_buffer_writer_version(void) {return(3);}

DLLEXPORT miBoolean p_buffer_writer(miColor *result, miState *state, p_buffer_writer_t *paras)
{
	if(state->type == miRAY_SHADOW || state->type == miRAY_DISPLACE)
		return miFALSE;

	miVector orig_normal = state->normal;
	miBoolean restore_normal = *mi_eval_boolean(&paras->restore_normal);

	if((state->type == miRAY_EYE) || (state->type == miRAY_TRANSPARENT))
	{
		if(*mi_eval_boolean(&paras->enable_fb0))
		{
			mi_fb_put(state, 0, mi_eval_color(&paras->buffer_0));
			if(restore_normal)
				state->normal = orig_normal;
		}

		if(*mi_eval_boolean(&paras->enable_fb1))
		{
			mi_fb_put(state, 1, mi_eval_color(&paras->buffer_1));
			if(restore_normal)
				state->normal = orig_normal;
		}

		if(*mi_eval_boolean(&paras->enable_fb2))
		{
			mi_fb_put(state, 2, mi_eval_color(&paras->buffer_2));
			if(restore_normal)
				state->normal = orig_normal;
		}

		if(*mi_eval_boolean(&paras->enable_fb3))
		{
			mi_fb_put(state, 3, mi_eval_color(&paras->buffer_3));
			if(restore_normal)
				state->normal = orig_normal;
		}

		if(*mi_eval_boolean(&paras->enable_fb4))
		{
			mi_fb_put(state, 4, mi_eval_color(&paras->buffer_4));
			if(restore_normal)
				state->normal = orig_normal;
		}

		if(*mi_eval_boolean(&paras->enable_fb5))
		{
			mi_fb_put(state, 5, mi_eval_color(&paras->buffer_5));
			if(restore_normal)
				state->normal = orig_normal;
		}

		if(*mi_eval_boolean(&paras->enable_fb6))
		{
			mi_fb_put(state, 6, mi_eval_color(&paras->buffer_6));
			if(restore_normal)
				state->normal = orig_normal;
		}

		if(*mi_eval_boolean(&paras->enable_fb7))
		{
			mi_fb_put(state, 7, mi_eval_color(&paras->buffer_7));
			if(restore_normal)
				state->normal = orig_normal;
		}
	}
	*result = *mi_eval_color(&paras->render);
	return(miTRUE);
}
}
