//  Copyright (c) 2005, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <math.h>
#include "shader.h"

typedef struct
{
	miScalar	color_R;
	miScalar	color_G;
	miScalar	color_B;
	miScalar	color_A;
	miBoolean	clamp_black;
	miBoolean	clamp_white;
}p_float4_to_color_t;

extern "C"
{
DLLEXPORT int p_float4_to_color_version(void) {return(2);}

DLLEXPORT miBoolean p_float4_to_color(miColor *result, miState *state, p_float4_to_color_t *paras)
{
	result->r = *mi_eval_scalar(&paras->color_R);
	result->g = *mi_eval_scalar(&paras->color_G);
	result->b = *mi_eval_scalar(&paras->color_B);
	result->a = *mi_eval_scalar(&paras->color_A);

	if(*mi_eval_boolean(&paras->clamp_black))
	{
		if(result->r < 0)
			result->r = 0;
		if(result->g < 0)
			result->g = 0;
		if(result->b < 0)
			result->b = 0;
		if(result->a < 0)
			result->a = 0;
	}
	if(*mi_eval_boolean(&paras->clamp_white))
	{
		if(result->r > 1)
			result->r = 1;
		if(result->g > 1)
			result->g = 1;
		if(result->b > 1)
			result->b = 1;
		if(result->a > 1)
			result->a = 1;
	}
	return(miTRUE);
}
}
