#include <math.h>
#include "shader.h"

typedef struct
{
	float r;
	float g;
	float b;
}maColor;

struct maya_out
{
	miColor outColor;
	miColor outGlowColor;
	miColor outMatteOpacity;
	miColor outTransparency;
};

typedef struct
{
	miColor		inColor;
	miBoolean	useNormal;
	miVector	inputNormal;
	int			matteOpacityMode;
	miScalar	matteOpacity;
}p_maya_shadingengine_t;

extern "C"
{
DLLEXPORT int p_maya_shadingengine_version(void) {return(2);}

DLLEXPORT miBoolean p_maya_shadingengine(maya_out *result, miState *state, p_maya_shadingengine_t *paras)
{
	if(state->type == miRAY_SHADOW || state->type == miRAY_DISPLACE)
		return miFALSE;

	miVector *bump_normal;
	if(*mi_eval_boolean(&paras->useNormal))													//Change normal(bump)
	{
		bump_normal = mi_eval_vector(&paras->inputNormal);
		mi_vector_normalize(bump_normal);
		state->normal = *bump_normal;
		state->dot_nd = mi_vector_dot(bump_normal, &state->dir);		
	}

	miColor		*input = mi_eval_color(&paras->inColor);
	int			matteOpacityMode = *mi_eval_integer(&paras->matteOpacityMode);
	miScalar	matteOpacity = *mi_eval_scalar(&paras->matteOpacity);

	result->outColor = *input;
	result->outGlowColor.r = result->outGlowColor.g = result->outGlowColor.b = 0;


	if((state->type == miRAY_EYE) || (state->type == miRAY_TRANSPARENT))
	{
		switch(matteOpacityMode)
		{
			case 0:			//black hole
				result->outColor.r = result->outColor.g = result->outColor.b = 0;
				result->outMatteOpacity.r = result->outMatteOpacity.g = result->outMatteOpacity.b = 0;
				result->outTransparency.r = result->outTransparency.g = result->outTransparency.b = 0; 
				break;
			case 1:			//solid matte
				result->outMatteOpacity.r = result->outMatteOpacity.g = result->outMatteOpacity.b = matteOpacity;
				result->outTransparency.r = result->outTransparency.g = result->outTransparency.b = 0; 
				break;
			case 2:			//opacity gain
				result->outMatteOpacity.r = result->outMatteOpacity.g = result->outMatteOpacity.b = input->a * matteOpacity;
				result->outTransparency.r = result->outTransparency.g = result->outTransparency.b = 1 -input->a; 
				break;
			default:		//opacity gain
				result->outMatteOpacity.r = result->outMatteOpacity.g = result->outMatteOpacity.b = input->a * matteOpacity;
				result->outTransparency.r = result->outTransparency.g = result->outTransparency.b = 1 -input->a; 
				break;
		}
	}
	else			//reflection, refraction and others
	{
		if(matteOpacityMode)
		{
				result->outMatteOpacity.r = result->outMatteOpacity.g = result->outMatteOpacity.b = matteOpacity;
				result->outTransparency.r = result->outTransparency.g = result->outTransparency.b = 0; 
		}
		else			//black hole	
		{
				result->outColor.r = result->outColor.g = result->outColor.b = 0;
				result->outMatteOpacity.r = result->outMatteOpacity.g = result->outMatteOpacity.b = 0;
				result->outTransparency.r = result->outTransparency.g = result->outTransparency.b = 0; 
		}
	}
	return(miTRUE);
}
}
