//  Copyright (c) 2005, Pavel Ledin
//
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//  *       Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  *       Redistributions in binary form must reproduce the above
//  copyright notice, this list of conditions and the following disclaimer
//  in the documentation and/or other materials provided with the
//  distribution.
//  *       Neither the name of Pavel Ledin nor the names of
//  its other contributors may be used to endorse or promote products derived
//  from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
//  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
//  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
//  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <math.h>
#include "shader.h"

typedef struct
{
	miBoolean	enable_reflection;
	int			reflection_depth;
	miBoolean	enable_refraction;
	int			refraction_depth;
	miBoolean	enable_transparency;
	int			transparency_depth;
	miBoolean	enable_finalgather;
	miBoolean	enable_environment;
	miBoolean	enable_shadow;
	miColor		eye;
	miColor		reflection;
	miColor		refraction;
	miColor		transparency;
	miColor		finalgather;
	miColor		environment;
	miColor		shadow;
}p_ray_type_t;

extern "C"
{
DLLEXPORT int p_ray_type_version(void) {return(3);}

DLLEXPORT miBoolean p_ray_type(miColor *result, miState *state,	p_ray_type_t *paras)
{
	if(state->type == miRAY_SHADOW || state->type == miRAY_DISPLACE)
		return miFALSE;

	switch(state->type)
	{
		case miRAY_EYE:
			*result = *mi_eval_color(&paras->eye);
			break;
		case miRAY_REFLECT:
			if(*mi_eval_boolean(&paras->enable_reflection) && state->reflection_level >= *mi_eval_boolean(&paras->reflection_depth))
				*result = *mi_eval_color(&paras->reflection);
			else
				*result = *mi_eval_color(&paras->eye);
			break;
		case miRAY_REFRACT:
			if(*mi_eval_boolean(&paras->enable_refraction) && state->refraction_level >= *mi_eval_boolean(&paras->refraction_depth))
				*result = *mi_eval_color(&paras->refraction);
			else
				*result = *mi_eval_color(&paras->eye);
			break;
		case miRAY_TRANSPARENT:
			if(*mi_eval_boolean(&paras->enable_transparency) && state->refraction_level >= *mi_eval_boolean(&paras->transparency_depth))
				*result = *mi_eval_color(&paras->transparency);
			else
				*result = *mi_eval_color(&paras->eye);
			break;
		case miRAY_FINALGATHER:
			if(*mi_eval_boolean(&paras->enable_finalgather) == miTRUE)
				*result = *mi_eval_color(&paras->finalgather);
			else
				*result = *mi_eval_color(&paras->eye);
			break;
		case miRAY_ENVIRONMENT:
			if(*mi_eval_boolean(&paras->enable_environment) == miTRUE)
				*result = *mi_eval_color(&paras->environment);
			else
				*result = *mi_eval_color(&paras->eye);
			break;
		case miRAY_SHADOW:
			if(*mi_eval_boolean(&paras->enable_shadow) == miTRUE)
				*result = *mi_eval_color(&paras->shadow);
			else
				*result = *mi_eval_color(&paras->eye);
			break;
		default:
			*result = *mi_eval_color(&paras->eye);
			break;
	}
	return(miTRUE);
}
}
